# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np

"""
    Fills data within margin(s) where data may not be correct.
    
    Required Arguments:
        tvar : str
            Name of tvar to modify.  
        margin : list
            Index values between which data values will be modified
        dt : int/float
            Step size to interpolate rows, then modify
    
    Optional Arguments:
        func : str
            Either 'nan' or 'ffill', which overrides normal interpolation with NaN
            substitution or forward-filled values.
    Returns:
        None
    
    Examples:
        >>> Null data between [3,7], and [13,17], with step size 2.3.
        >>> pytplot.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1],[2,2],[100,100],[4,4],[5,5],[6,6],[7,7]]})
        >>> pytplot.degap('d',[[3,7],[13,17]],2.3,func='nan')  
    """


def degap(tvar,dt):
    #Not yet implemented, see below for previous code
    pass
    '''
    tv = pytplot.data_quants[tvar].data.copy()
    tvi = pytplot.data_quants[tvar].data.index.copy().tolist()
    try:
        tspec = pytplot.data_quants[tvar].spec_bins.copy()
    except:
        pass
    
    print(tvi)
    d = np.roll(tvi,1)
    print(d)
    print(tvi)
    shift_arr = abs(d - np.array(tvi))
    delta = np.where(abs(d - np.array(tvi))>dt)
    
    #for i in len(tv.columns):
    #    for d_i,val in enumerate(shift_arr):
    #        pass
#degap('d',3)
    
# def degap(tvar,margin,dt,func=None):
#     tv1 = pytplot.data_quants[tvar].data.copy()
#     indices = []
#     #create array of dt between margin parameters
#     for m in margin:
#         if type(m) == int:
#             new_index = np.arange(margin[0],margin[1],dt)
#         else:
#             new_index = np.arange(m[0],m[1],dt)
#         new_index = np.append(new_index,new_index[-1]+dt)
#         indices = indices + [new_index]
#     #add any new indices to current dataframe indices
#     new_index = np.unique(np.append(indices,tv1.index))
#     #interpolate to create data for new values
#     pytplot.resample(tvar,new_index,tvar+'_interp')
#         
#     if func == 'nan':
#         tv1 = pytplot.data_quants[tvar+'_interp'].data
#         #any location between margins will be changed to NaN
#         for i in indices:
#             tv1.loc[i[0]:i[-1]] = np.nan
#         if (pytplot.data_quants[tvar].spec_bins is not None) and (pytplot.data_quants[tvar].spec_bins_time_varying == True):
#             pytplot.store_data(tvar+'_nan',data = {'x':tv1.index.values,'y':tv1,'v':pytplot.data_quants[tvar].spec_bins})
#         else:
#             pytplot.store_data(tvar+'_nan',data = {'x':tv1.index.values,'y':tv1})
#         pytplot.del_data(name=tvar+'_interp')
# 
#     if func == 'ffill':
#         tv1 = pytplot.data_quants[tvar+'_interp'].data
#         #any location between margins will be changed to NaN
#         for i in indices:
#             tv1.loc[i[0]:i[-1]] = np.nan
#         #forward fill NaNs in the dataframe
#         tv1 = tv1.fillna(method='ffill')
#         if (pytplot.data_quants[tvar].spec_bins is not None) and (pytplot.data_quants[tvar].spec_bins_time_varying == True):
#             pytplot.store_data(tvar+'_ffill',data = {'x':tv1.index.values,'y':tv1,'v':pytplot.data_quants[tvar].spec_bins})
#         else:
#             pytplot.store_data(tvar+'_ffill',data = {'x':tv1.index.values,'y':tv1})
#         pytplot.del_data(name=tvar+'_interp')
# 
#     return
#           
    '''


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np

"""
    Change specified 'flagged' data to NaN.
    
    Parameters:
        tvar1 : str
            Name of tvar to use for data clipping.  
        flags : int,list
            Flagged data will be converted to NaNs.
        newtvar : str
            Name of new tvar for deflagged data storage
            
    Returns:
        None
    
    Examples:
        >>> Remove any instances of [100,90,7,2,57] from 'd', store in 'e'.
        >>> pytplot.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1],[2,2],[100,4],[4,90],[5,5],[6,6],[7,7]]})
        >>> pytplot.deflag('d',[100,90,7,2,57],'e')
    """

                              
def deflag(tvar1,flags,newtvar='tvar_deflag'):
    #if int input, make list
    if not isinstance(flags,list):
        flags = [flags]
    #grab column indices
    df_index = pytplot.data_quants[tvar1].data.columns
    new_df = []
    tvar_orig = pytplot.data_quants[tvar1].data.copy()
    #for each column of dataframe
    for i in df_index:
        tv2_col = [item[i] for item in tvar_orig.values]
        for j,valj in enumerate(tv2_col):
            #if one of flagged values, convert to NaN
            if valj in flags:
                tv2_col[j] = np.NaN
        new_df = new_df + [tv2_col]
    new_df = np.transpose((list(new_df)))
    #store deflagged tvar
    if (pytplot.data_quants[tvar1].spec_bins is not None) and (pytplot.data_quants[tvar1].spec_bins_time_varying == True):
        pytplot.store_data(newtvar, data={'x':tvar_orig.index,'y':new_df,'v':pytplot.data_quants[tvar1].spec_bins})
    else:
        pytplot.store_data(newtvar, data={'x':tvar_orig.index,'y':new_df})
    return



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np

#PARTIAL ADD ACROSS COLUMNS
#add tvar data across specific columns, store in new_tvar
def add_across_partial(tvar1,column_range,new_tvar='tvar_aap'):
    #separate and add data
    time = pytplot.data_quants[tvar1].data.index.copy()
    data1 = pytplot.data_quants[tvar1].data.copy()
    data = []
    #grab column data
    if len(column_range)==2 and isinstance(column_range[0],int):
        range_start = column_range[0]
        range_end = column_range[1]
        add_col = list(range(range_start,range_end+1))
        datasum = data1[add_col].sum(axis=1)
        data = data + [list(datasum)]
    else:
        for i in column_range:
            #if not a list
            if type(i) == int:
                data = data + [list(data1[i])]
            #sum across listed column range
            else:
                range_start = i[0]
                range_end = i[1]
                add_col = list(range(range_start,range_end+1))
                datasum = data1[add_col].sum(axis=1)
                data = data + [list(datasum)]
    #store added data
    pytplot.store_data(new_tvar,data={'x':time, 'y':np.transpose(data)})
    return new_tvar



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pytplot
from . import tplot_utilities

def tplot_options(option, value):
    """
    This function allows the user to set several global options for the generated plots.  
    
    Parameters:
        option : str
            The name of the option.  See section below  
        value : str/int/float/list
            The value of the option.  See section below.  
            
    Options:
        ============  ==========   =====
        Options       Value type   Notes
        ============  ==========   =====
        title         str          Title of the the entire output
        title_size    int          Font size of the output 
        wsize         [int, int]   [height, width], pixel size of the plot window
        title_align   int          Offset position in pixels of the title   
        var_label     srt          Name of the tplot variable to be used as another x axis
        alt_range     [flt, flt]   The min and max altitude to be plotted on all alt plots
        map_x_range   [int, int]   The min and max longitude to be plotted on all map plots
        map_y_range   [int, int]   The min and max latitude to be plotted on all map plots
        x_range       [flt, flt]   The min and max x_range (usually time) to be plotted on all Spec/1D plots
        crosshair     bool         Option allowing crosshairs and crosshair legend
        roi           [str, str]   Times between which there's a region of interest for a user
        ============  ==========   =====
    
    Returns:
        None
    
    Examples:
        >>> # Set the plot title
        >>> import pytplot
        >>> pytplot.tplot_options('title', 'SWEA Data for Orbit 1563')
        
        >>> # Set the window size 
        >>> pytplot.tplot_options('wsize', [1000,500])
        
    
    """
    
    
    
    option = option.lower()
    
    temp = tplot_utilities.set_tplot_options(option, value, pytplot.tplot_opt_glob)
    pytplot.tplot_opt_glob = temp
    
    return



import pytplot

#SPLIT TVAR
#store columns of TVar into new TVars
def split_vec(tvar,newtvars=None,columns='all'):
    #separate and add data
    time = pytplot.data_quants[tvar].data.index
    data = pytplot.data_quants[tvar].data
    df = pytplot.data_quants[tvar]
    defaultlist = []
    #grab column data
    if columns == 'all':
        columns = pytplot.data_quants[tvar].data.columns.values
    for i,val in enumerate(columns):
        #if not a list
        if isinstance(val,int):
            range_start = val
            range_end = val
        else:
            range_start = val[0]
            range_end = val[1]
        split_col = list(range(range_start,range_end+1))
        #store split data
        defaultname = 'data_' + str(i)
        defaultlist = defaultlist + [defaultname]
        if newtvars is None:
            pytplot.store_data(defaultname,data={'x':time, 'y':data[split_col]})
        else:
            pytplot.store_data(newtvars[i],data={'x':time, 'y':data[split_col]})
    if newtvars is None:
        return defaultlist
    else:
        return newtvars



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

from pytplot import data_quants

def get_data(name):
    """
    This function extracts the data from the Tplot Variables stored in memory.  
    
    Parameters:
        name : str 
            Name of the tplot variable
         
    Returns:
        time_val : pandas dataframe index
        data_val : list
            
    Examples:
        >>> # Retrieve the data from Variable 1
        >>> import pytplot
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pytplot.store_data("Variable1", data={'x':x_data, 'y':y_data})
        >>> time, data = pytplot.get_data("Variable1")

    """
    
    global data_quants
    if name not in data_quants.keys():
        print("That name is currently not in pytplot")
        return
    
    temp_data_quant = data_quants[name]
    data_val = temp_data_quant.data.values
    time_val = temp_data_quant.data.index
    
    return(time_val, data_val)


import pyqtgraph as pg
import numpy as np
import pytplot
from pytplot import tplot_utilities
from pyqtgraph.Qt import QtGui, QtCore


def interactiveplot(t_average=None):
    """ If the interactive option is set to True in tplot, this function will take in the stored tplot variables
    and create a 2D interactive window that will pop up when any one of the tplot variables is plotted (so long
    as at least one of the tplot variables is a spectrogram). If the mouse hovers over a spectrogram plot, data
    for that point in time on the spectrogram plot will be plotted in the 2D interactive window. If the mouse
    hovers over a non-spectrogram plot, the 2D interactive window returns an empty plot. If the 't_average'
    option is selected, then the interactive window's y values will be time-averaged values, where the
    amount of time for which those values have been averaged is determined by the number of seconds the user
    indicates. """

    # Grab names of data loaded in as tplot variables.
    names = list(pytplot.data_quants.keys())
    # Get data we'll actually work with here.
    valid_variables = tplot_utilities.get_data(names)

    # Don't plot anything unless we have spectrograms with which to work.
    if valid_variables:
        # Get z label
        labels = tplot_utilities.get_labels_axis_types(names)

        # Put together data in easy-to-access format for plots.
        data = {}
        for name in valid_variables:
            bins = tplot_utilities.get_bins(name)
            time_values, z_values = tplot_utilities.get_z_t_values(name)
            data[name] = [bins, z_values, time_values]

        # Set up the 2D interactive plot
        pytplot.interactive_window = pg.GraphicsWindow()
        pytplot.interactive_window.resize(1000, 600)
        pytplot.interactive_window.setWindowTitle('Interactive Window')
        plot = pytplot.interactive_window.addPlot(title='2D Interactive Plot', row=0, col=0)
        # Make it so that whenever this first starts up, you just have an empty plot
        plot_data = plot.plot([], [])

        # The following update function is passed to change_hover_time in the HoverTime class
        # defined in __init__.py. For reference, "t" essentially originates inside of
        # TVarFigure(1D/Spec/Alt/Map), inside the _mousemoved function. It calls
        # "self._mouseMovedFunction(int(mousePoint.x()))" and that is called every time the mouse is
        # moved by Qt. Therefore, it gives the location of the mouse on the x axis. In tplot,
        # mouse_moved_event is set to pytplot.hover_time.change_hover_time, so the mouseMovedFunction
        # is pytplot.hover_time.change_hover_time. Thus, whenever change_hover_time is called, it
        # calls every other function that is registered. Since the below function update() is
        # registered as a listener, it'll update whenever hover_time is updated.
        # to the HoverTime class with "t" as the input.

        # TL;DR, t comes from getting the mouse location in pyqtgraph every time the mouse is moved
        # and the below function will update the plot's position as the mouse is moved.
        def update(t, name):
            if name in valid_variables:
                # Get the time closest to the x position the mouse is over.
                time_array = np.array(data[name][2])
                array = np.asarray(time_array)
                idx = (np.abs(array - t)).argmin()

                # If user indicated they wanted the interactive plot's axes to be logged, log 'em.
                # But first make sure that values in x and y are loggable!
                x_axis = False
                y_axis = False
                # Checking x axis
                if np.nanmin(data[name][0][:]) < 0:
                    print('Negative data is incompatible with log plotting.')
                elif np.nanmin(data[name][0][:]) >= 0 and labels[name][2] == 'log':
                    x_axis = True
                # Checking y axis
                if np.nanmin(list(data[name][1][idx])) < 0:
                    print('Negative data is incompatible with log plotting')
                elif np.nanmin(list(data[name][1][idx])) >= 0 and labels[name][3] == 'log':
                    y_axis = True

                # Set plot labels
                plot.setLabel('bottom', '{}'.format(labels[name][0]))
                plot.setLabel('left', '{}'.format(labels[name][1]))
                plot.setLogMode(x=x_axis, y=y_axis)
                # Update x and y range if user modified it
                tplot_utilities.set_x_range(name, x_axis, plot)
                tplot_utilities.set_y_range(name, y_axis, plot)

                if 't_average' in pytplot.data_quants[name].extras:
                    # If the user indicated that they wanted to average the interactive plot's y values based on a
                    # certain time range around the cursor location, we then want to get average of values around
                    # the cursor location.
                    t_min = data[name][2][0]
                    t_max = data[name][2][-1]

                    delta = pytplot.data_quants[name].extras['t_average']/int(2)

                    left_bound = data[name][2][idx] - delta
                    right_bound = data[name][2][idx] + delta

                    if (left_bound - t_min >= 0) and (t_max - right_bound >= 0):
                        # Find index of left and right bounds, no fancy foot work necessary.
                        idx_left = (np.abs(array - left_bound)).argmin()
                        idx_right = (np.abs(array - right_bound)).argmin()
                    elif left_bound - t_min < 0:
                        # Find the number of seconds difference between the cursor's
                        # left bound and the minimum time in the dataset, add that
                        # difference to the right bound time (since you want to push the bound
                        # forward in time, and set the left bound's index to be 0.
                        idx_left = 0
                        diff = right_bound + (t_min - left_bound)
                        idx_right = (np.abs(array - diff)).argmin()
                    elif t_max - right_bound < 0:
                        # Find the number of seconds difference between the cursor's
                        # right bound and the maximum time in the dataset, subtract that
                        # difference from the left bound time (since you want to push the bound
                        # back in time), and set the right bound's index to be -1.
                        idx_right = -1
                        diff = left_bound - (right_bound - t_max)
                        idx_left = (np.abs(array - diff)).argmin()
                    elif (left_bound - t_min < 0) and (t_max - right_bound < 0):
                        # The user is asking to average the entire time frame of the dataset...
                        # dunno why they want that, but if they do, use the time-averaged static plot,
                        # not this.
                        print(
                            'This plot isn\'t appropriate for what you want, use the time-averaged static plot.')

                    # Average values based on the calculated right and left bounds' indices.
                    time_diff = abs(idx_right - idx_left)
                    # Make sure to account for edge problem
                    if idx_right != -1:
                        y_values_slice = data[name][1][idx_left:idx_right + 1]
                    else:
                        y_values_slice = data[name][1][idx_left:]
                    y_values_avgd = np.sum(y_values_slice, axis=0)/np.float(time_diff)

                    # Update x and y range if user modified it
                    tplot_utilities.set_x_range(name, x_axis, plot)
                    tplot_utilities.set_y_range(name, y_axis, plot)

                    try:
                        # Plot data based on time we're hovering over
                        plot_data.setData(data[name][0][:], y_values_avgd)
                    except ZeroDivisionError:
                        pass
                else:
                    # Update x and y range if user modified it
                    tplot_utilities.set_x_range(name, x_axis, plot)
                    tplot_utilities.set_y_range(name, y_axis, plot)
                    # If the user just wants a plain jane interactive plot...
                    # Plot data based on time we're hovering over'
                    try:
                        plot_data.setData(data[name][0][:], list(data[name][1][idx]))
                    except ZeroDivisionError:
                        pass

            else:
                # Cover the situation where you hover over a non-spectrogram plot.
                plot.setLogMode(False, False)
                plot.setLabel('bottom', '')
                plot.setLabel('left', '')
                plot_data.setData([], [])

        # Make the above function called whenever hover_time is updated.
        pytplot.hover_time.register_listener(update)

        # Start Qt event loop unless running in interactive mode.
        #import sys
        #if not (hasattr(sys, 'ps1')) or not hasattr(QtCore, 'PYQT_VERSION'):
        #    QtGui.QApplication.instance().exec_()



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

from __future__ import division
import datetime
import pandas as pd
import numpy as np

from pytplot import data_quants, TVar
from .del_data import del_data
import pytplot

tplot_num = 1


def store_data(name, data=None, delete=False, newname=None):
    
    """
    This function creates a "Tplot Variable" based on the inputs, and
    stores this data in memory.  Tplot Variables store all of the information
    needed to generate a plot.  
    
    Parameters:
        name : str 
            Name of the tplot variable that will be created
        data : dict
            A python dictionary object.  
            
            'x' should be a 1-dimensional array that represents the data's x axis.  Typically this data is time,
            represented in seconds since epoch (January 1st 1970)
            
            'y' should be the data values. This can be 2 dimensions if multiple lines or a spectrogram are desired.
            
            'v' is optional, and is only used for spectrogram plots.  This will be a list of bins to be used.  If this
            is provided, then 'y' should have dimensions of x by z.
            
            'x' and 'y' can be any data format that can be read in by the pandas module.  Python lists, numpy arrays,
            or any pandas data type will all work.
        delete : bool, optional
            Deletes the tplot variable matching the "name" parameter
        newname: str
            Renames TVar to new name
        
    .. note::
        If you want to combine multiple tplot variables into one, simply supply the list of tplot variables to the
        "data" parameter.  This will cause the data to overlay when plotted.
        
    Returns:
        None
        
    Examples:
        >>> # Store a single line
        >>> import pytplot
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pytplot.store_data("Variable1", data={'x':x_data, 'y':y_data})
    
        >>> # Store a two lines
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [[1,5],[2,4],[3,3],[4,2],[5,1]]
        >>> pytplot.store_data("Variable2", data={'x':x_data, 'y':y_data})
        
        >>> # Store a spectrogram
        >>> x_data = [1,2,3]
        >>> y_data = [ [1,2,3] , [4,5,6], [7,8,9] ]
        >>> v_data = [1,2,3]
        >>> pytplot.store_data("Variable3", data={'x':x_data, 'y':y_data, 'v':v_data})
        
        >>> # Combine two different line plots
        >>> pytplot.store_data("Variable1and2", data=['Variable1', 'Variable2'])
        
        >>> #Rename TVar
        >>> pytplot.store_data('a', data={'x':[0,4,8,12,16], 'y':[1,2,3,4,5]})
        >>> pytplot.store_data('a',newname='f')

    """
    
    global tplot_num
    create_time = datetime.datetime.now()
    
    if delete is True:
        del_data(name)
        return

    if data is None and newname is None:
        print('Please provide data.')
        return
    
    if newname is not None:
        pytplot.tplot_rename(name, newname)
        return
    
    if isinstance(data, list):
        base_data = get_base_tplot_vars(data)
        # Use first tplot var as the time range
        trange = [np.nanmin(data_quants[base_data[0]].data.index), 
                  np.nanmax(data_quants[base_data[0]].data.index)]
        df = base_data
        spec_bins = None
    else:             
        df = format_ydata(data['y'])            
        times = data['x']
        if len(times) != len(df.index):
            print("The lengths of x and y do not match!")
            return
        elif isinstance(times, pd.Series):
            df = df.set_index(data['x'])
        else:
            df['Index'] = times
            df = df.set_index('Index', drop=True)
        trange = [np.nanmin(times), np.nanmax(times)]
        
        if 'v' in data or 'v2' in data:
            # Generally the data is 1D, but occasionally
            # the bins will vary in time.
            if 'v' in data:
                spec_bins = data['v']
            else:
                spec_bins = data['v2']
            
            if type(spec_bins) is not pd.DataFrame:
                spec_bins = pd.DataFrame(spec_bins)
                if len(spec_bins.columns) != 1:
                    if len(spec_bins) == len(df.index):
                        spec_bins = spec_bins.set_index(df.index)  
                    else:
                        print("Length of v and x do not match.  Cannot create tplot variable.")
                        return 
                else:
                    spec_bins = spec_bins.transpose()         
        else:
            spec_bins = None
        
    xaxis_opt = dict(axis_label='Time')
    yaxis_opt = dict(axis_label=name) if (spec_bins is None) else dict(axis_label='')
    zaxis_opt = dict(axis_label='') if (spec_bins is None) else dict(axis_label=name)
    line_opt = {}
    dtype = ''
    time_bar = []
    # Dictionary to keep track of extra details needed for plotting
    #     that aren't actual attributes in Bokeh
    extras = dict(panel_size=1)
    links = {}
    temp = TVar(name, tplot_num, df, spec_bins, xaxis_opt, yaxis_opt, zaxis_opt, line_opt,
                trange, dtype, create_time, time_bar, extras, links)
    
    data_quants[name] = temp
    data_quants[name].yaxis_opt['y_range'] = get_y_range(df, spec_bins)
    
    return


def get_base_tplot_vars(data):
    base_vars = []
    if not isinstance(data, list):
        data = [data]
    for var in data:
        if isinstance(data_quants[var].data, list):
            base_vars += get_base_tplot_vars(data_quants[var].data)
        else:
            base_vars += [var]
    return base_vars


def get_y_range(data, spec_bins):
    # This is for the numpy RuntimeWarning: All-NaN axis encountered
    # with np.nanmin below
    import warnings
    warnings.filterwarnings("error")

    if spec_bins is not None:
        ymin = np.nanmin(spec_bins)
        ymax = np.nanmax(spec_bins)
        return [ymin, ymax]
    else:
        datasets = []
        y_min_list = []
        y_max_list = []
        if isinstance(data, list):
            for oplot_name in data:
                datasets.append(data_quants[oplot_name].data)
        else:
            datasets.append(data)
    
        for dataset in datasets:
            dataset_temp = dataset.replace([np.inf, -np.inf], np.nan)
            try:
                y_min_list.append(np.nanmin(dataset_temp.min(skipna=True).tolist()))
                y_max_list.append(np.nanmax(dataset_temp.max(skipna=True).tolist()))
            except RuntimeWarning:
                y_min_list.append(np.nan)
                y_max_list.append(np.nan)
        
        y_min = min(y_min_list)
        y_max = max(y_max_list)
        
        if y_min == y_max:
            # Show 10% and 10% below the straight line
            y_min = y_min-(.1*np.abs(y_min))
            y_max = y_max+(.1*np.abs(y_max))
        warnings.resetwarnings()
        return [y_min, y_max]


def format_ydata(data):
    # This function is not final, and will presumably change in the future
    # For 2D data, turn it into a Pandas dataframe
    # For 3D data, Sum over the second dimension, then turn into a Pandas dataframe
    # For 4D data, ignore the last dimension
    
    if data is not pd.DataFrame:
        matrix = np.array(data)
        if len(matrix.shape) > 2:
            matrix = np.nansum(matrix, 1)
        if len(matrix.shape) > 2:
            matrix = matrix[:, :, 0]
            
    else:
        return data
    
    return_data = pd.DataFrame(matrix)
    return return_data



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyqtgraph as pg
import numpy as np
import os
import pytplot
from pyqtgraph.Qt import QtCore
from .CustomAxis.BlankAxis import BlankAxis
from .CustomLegend.CustomLegend import CustomLegendItem
from .CustomAxis.AxisItem import AxisItem
from .CustomViewBox.NoPaddingPlot import NoPaddingPlot


class TVarFigureMap(pg.GraphicsLayout):
    def __init__(self, tvar_name, show_xaxis=False, mouse_function=None):
        self.tvar_name = tvar_name
        self.show_xaxis = show_xaxis
        self.crosshair = pytplot.tplot_opt_glob['crosshair']

        # Sets up the layout of the Tplot Object
        pg.GraphicsLayout.__init__(self)
        self.layout.setHorizontalSpacing(50)
        self.layout.setContentsMargins(0, 0, 0, 0)
        # Set up the x axis
        self.xaxis = pg.AxisItem(orientation='bottom')
        self.xaxis.setHeight(35)
        self.xaxis.enableAutoSIPrefix(enable=False)
        # Set up the y axis
        self.yaxis = AxisItem("left")
        self.yaxis.setWidth(100)

        vb = NoPaddingPlot()
        self.plotwindow = self.addPlot(row=0, col=0, axisItems={'bottom': self.xaxis, 'left': self.yaxis}, viewBox=vb)
        self.plotwindow.vb.setLimits(xMin=0, xMax=360, yMin=-90, yMax=90)

        # Set up the view box needed for the legends
        self.legendvb = pg.ViewBox(enableMouse=False)
        self.legendvb.setMaximumWidth(100)
        self.legendvb.setXRange(0, 1, padding=0)
        self.legendvb.setYRange(0, 1, padding=0)
        self.addItem(self.legendvb, 0, 1)

        self.curves = []
        self.colors = self._setcolors()
        self.colormap = self._setcolormap()

        self.labelStyle = {'font-size': str(pytplot.data_quants[self.tvar_name].extras['char_size'])+'pt'}

        if show_xaxis:
            self.plotwindow.showAxis('bottom')
        else:
            self.plotwindow.hideAxis('bottom')

        self._mouseMovedFunction = mouse_function

        self.label = pg.LabelItem(justify='left')
        self.addItem(self.label, row=1, col=0)

        # Set legend options
        self.hoverlegend = CustomLegendItem(offset=(0, 0))
        self.hoverlegend.setItem("Date: ", "0")
        self.hoverlegend.setItem("Time: ", "0")
        self.hoverlegend.setItem("Latitude:", "0")
        self.hoverlegend.setItem("Longitude:", "0")
        self.hoverlegend.setVisible(False)
        self.hoverlegend.setParentItem(self.plotwindow.vb)

    def _set_crosshairs(self):
        self.vLine = pg.InfiniteLine(angle=90, movable=False, pen=pg.mkPen('k'))
        self.hLine = pg.InfiniteLine(angle=0, movable=False, pen=pg.mkPen('k'))
        self.plotwindow.addItem(self.vLine, ignoreBounds=True)
        self.plotwindow.addItem(self.hLine, ignoreBounds=True)
        self.vLine.setVisible(False)
        self.hLine.setVisible(False)

    def buildfigure(self):
        self._setxrange()
        self._setyrange()
        self._setyaxistype()
        self._setzaxistype()
        self._setzrange()
        self._setbackground()
        self._visdata()
        self._setxaxislabel()
        self._setyaxislabel()
        self._addlegend()
        self._addtimebars()
        if self.crosshair:
            self._set_crosshairs()
            self._addmouseevents()

    def _setxaxislabel(self):
        self.xaxis.setLabel("Longitude", **self.labelStyle)

    def _setyaxislabel(self):
        self.yaxis.setLabel("Latitude", **self.labelStyle)

    def _setyaxislabel(self):
        self.yaxis.setLabel("Latitude")

    def getfig(self):
        return self

    def _visdata(self):
        datasets = []
        if isinstance(pytplot.data_quants[self.tvar_name].data, list):
            for oplot_name in pytplot.data_quants[self.tvar_name].data:
                datasets.append(pytplot.data_quants[oplot_name])
        else:
            datasets.append(pytplot.data_quants[self.tvar_name])

        cm_index = 0
        for dataset in datasets:
            t_link, lat = pytplot.get_data(dataset.links['lat'])
            lat = lat.transpose()[0]
            # Need to trim down the data points to fit within the link
            t_tvar = dataset.data.index.values
            data = dataset.data[0].values
            while t_tvar[-1] > t_link[-1]:
                t_tvar = np.delete(t_tvar, -1)
                data = np.delete(data, -1)
            while t_tvar[0] < t_link[0]:
                t_tvar = np.delete(t_tvar, 0)
                data = np.delete(data, 0)

            t_link, lon = pytplot.get_data(dataset.links['lon'])
            # Need to trim down the data points to fit within the link
            while t_tvar[-1] > t_link[-1]:
                t_tvar = np.delete(t_tvar, -1)
                data = np.delete(data, -1)
            while t_tvar[0] < t_link[0]:
                t_tvar = np.delete(t_tvar, 0)
                data = np.delete(data, 0)

            lon = lon.transpose()[0]
            for column_name in dataset.data.columns:
                values = data.tolist()
                colors = pytplot.tplot_utilities.get_heatmap_color(color_map=
                                                                   self.colormap[cm_index % len(self.colormap)],
                                                                   min_val=self.zmin, max_val=self.zmax, values=values,
                                                                   zscale=self.zscale)
                brushes = []
                for color in colors:
                    brushes.append(pg.mkBrush(color))
                self.curves.append(self.plotwindow.scatterPlot(lon.tolist(), lat.tolist(),
                                                               pen=pg.mkPen(None), brush=brushes))
                cm_index += 1

    def _setyaxistype(self):
        if self._getyaxistype() == 'log':
            self.plotwindow.setLogMode(y=True)
        else:
            self.plotwindow.setLogMode(y=False)
        return

    def _addlegend(self):
        zaxis = AxisItem('right')
        zaxis.setLabel(pytplot.data_quants[self.tvar_name].yaxis_opt['axis_label'], **self.labelStyle)

        if self.show_xaxis:
            emptyaxis = BlankAxis('bottom')
            emptyaxis.setHeight(35)
            p2 = self.addPlot(row=0, col=1, axisItems={'right': zaxis, 'bottom': emptyaxis}, enableMenu=False,
                              viewBox=self.legendvb)
        else:
            p2 = self.addPlot(row=0, col=1, axisItems={'right': zaxis}, enableMenu=False, viewBox=self.legendvb)
            p2.hideAxis('bottom')

        p2.buttonsHidden = True
        p2.setMaximumWidth(100)
        p2.showAxis('right')
        p2.hideAxis('left')
        colorbar = pg.ImageItem()
        colorbar.setImage(np.array([np.linspace(1, 2, 200)]).T)

        p2.addItem(colorbar)
        p2.setLogMode(y=(self.zscale == 'log'))
        p2.setXRange(0, 1, padding=0)
        if self.zscale == 'log':
            colorbar.setRect(QtCore.QRectF(0, np.log10(self.zmin), 1, np.log10(self.zmax) - np.log10(self.zmin)))
            # I have literally no idea why this is true, but I need to set the range twice
            p2.setYRange(np.log10(self.zmin), np.log10(self.zmax), padding=0)
            p2.setYRange(np.log10(self.zmin), np.log10(self.zmax), padding=0)
        else:
            colorbar.setRect(QtCore.QRectF(0, self.zmin, 1, self.zmax - self.zmin))
            p2.setYRange(self.zmin, self.zmax, padding=0)
        colorbar.setLookupTable(self.colormap[0])

    def _addmouseevents(self):
        if self.plotwindow.scene() is not None:
            self.plotwindow.scene().sigMouseMoved.connect(self._mousemoved)

    def _mousemoved(self, evt):
        # get current position
        pos = evt
        # if plot window contains position
        if self.plotwindow.sceneBoundingRect().contains(pos):
            mousepoint = self.plotwindow.vb.mapSceneToView(pos)
            # grab x and y mouse locations
            index_x = round(float(mousepoint.x()), 2)
            index_y = round(float(mousepoint.y()), 2)
            # get latitude and longitude arrays
            datasets = []
            if isinstance(pytplot.data_quants[self.tvar_name].data, list):
                for oplot_name in pytplot.data_quants[self.tvar_name].data:
                    datasets.append(pytplot.data_quants[oplot_name])
            else:
                datasets.append(pytplot.data_quants[self.tvar_name])

            time, latitude = pytplot.get_data(datasets[0].links['lat'])
            latitude = latitude.transpose()[0]
            time, longitude = pytplot.get_data(datasets[0].links['lon'])
            longitude = longitude.transpose()[0]
            # find closest time point to cursor
            radius = np.sqrt((latitude - index_y) ** 2 + (longitude - index_x) ** 2).argmin()
            time_point = time[radius]
            # get date and time
            date = (pytplot.tplot_utilities.int_to_str(time_point))[0:10]
            time = (pytplot.tplot_utilities.int_to_str(time_point))[11:19]

            # add crosshairs
            if self._mouseMovedFunction is not None:
                self._mouseMovedFunction(int(mousepoint.x()))
                self.vLine.setVisible(True)
                self.hLine.setVisible(True)
                self.vLine.setPos(mousepoint.x())
                self.hLine.setPos(mousepoint.y())

            # Set legend options
            self.hoverlegend.setVisible(True)
            self.hoverlegend.setItem("Date: ", date)
            self.hoverlegend.setItem("Time: ", time)
            self.hoverlegend.setItem("Longitude:", str(index_x))
            self.hoverlegend.setItem("Latitude:", str(index_y))
        else:
            self.hoverlegend.setVisible(False)
            self.vLine.setVisible(False)
            self.hLine.setVisible(False)

    def _getyaxistype(self):
        return 'linear'

    def _setzaxistype(self):
        if self._getzaxistype() == 'log':
            self.zscale = 'log'
        else:
            self.zscale = 'linear'

    def _getzaxistype(self):
        if 'z_axis_type' in pytplot.data_quants[self.tvar_name].zaxis_opt:
            return pytplot.data_quants[self.tvar_name].zaxis_opt['z_axis_type']
        else:
            return 'linear'

    def _setcolors(self):
        if 'line_color' in pytplot.data_quants[self.tvar_name].extras:
            return pytplot.data_quants[self.tvar_name].extras['line_color']
        else:
            return pytplot.tplot_utilities.rgb_color(['k', 'r', 'seagreen', 'b', 'darkturquoise', 'm', 'goldenrod'])

    def _setcolormap(self):
        colors = []
        if 'colormap' in pytplot.data_quants[self.tvar_name].extras:
            for cm in pytplot.data_quants[self.tvar_name].extras['colormap']:
                colors.append(pytplot.tplot_utilities.return_lut(cm))
            return colors
        else:
            return [pytplot.tplot_utilities.return_lut("inferno")]

    def getaxistype(self):
        axis_type = 'lat'
        link_y_axis = True
        return axis_type, link_y_axis

    def _setxrange(self):
        # Check if x range is set. Otherwise, set it.
        if 'map_x_range' in pytplot.tplot_opt_glob:
            self.plotwindow.setXRange(pytplot.tplot_opt_glob['map_x_range'][0],
                                      pytplot.tplot_opt_glob['map_x_range'][1])
        else:
            self.plotwindow.setXRange(0, 360)

    def _setyrange(self):
        # Check if y range is set.  Otherwise, y range is automatic
        if 'map_y_range' in pytplot.tplot_opt_glob:
            self.plotwindow.setYRange(pytplot.tplot_opt_glob['map_y_range'][0],
                                      pytplot.tplot_opt_glob['map_y_range'][1])
        else:
            self.plotwindow.vb.setYRange(-90, 90)

    def _setzrange(self):
        # Get Z Range
        if 'z_range' in pytplot.data_quants[self.tvar_name].zaxis_opt:
            self.zmin = pytplot.data_quants[self.tvar_name].zaxis_opt['z_range'][0]
            self.zmax = pytplot.data_quants[self.tvar_name].zaxis_opt['z_range'][1]
        else:
            if isinstance(pytplot.data_quants[self.tvar_name].data, list):
                # Check the first one
                dataset_temp = pytplot.data_quants[pytplot.data_quants[self.tvar_name].data[0]].data.replace(
                    [np.inf, -np.inf], np.nan)
            else:
                dataset_temp = pytplot.data_quants[self.tvar_name].data.replace([np.inf, -np.inf], np.nan)
            self.zmax = dataset_temp.max().max()
            self.zmin = dataset_temp.min().min()

            # Cannot have a 0 minimum in a log scale
            if self.zscale == 'log':
                zmin_list = []
                for column in pytplot.data_quants[self.tvar_name].data.columns:
                    series = pytplot.data_quants[self.tvar_name].data[column]
                    zmin_list.append(series.iloc[series.nonzero()[0]].min())
                self.zmin = min(zmin_list)

    def _addtimebars(self):
        # initialize dataset variable
        datasets = []
        # grab tbardict
        tbardict = pytplot.data_quants[self.tvar_name].time_bar
        ltbar = len(tbardict)
        # make sure data is in list format
        if isinstance(pytplot.data_quants[self.tvar_name].data, list):
            for oplot_name in pytplot.data_quants[self.tvar_name].data:
                datasets.append(pytplot.data_quants[oplot_name])
        else:
            datasets.append(pytplot.data_quants[self.tvar_name])
        for dataset in datasets:
            # for location in tbar dict
            for i in range(ltbar):
                # get times, color, point size
                test_time = pytplot.data_quants[self.tvar_name].time_bar[i]["location"]
                color = pytplot.data_quants[self.tvar_name].time_bar[i]["line_color"]
                pointsize = pytplot.data_quants[self.tvar_name].time_bar[i]["line_width"]
                # correlate given time with corresponding lat/lon points
                time, latitude = pytplot.get_data(dataset.links['lat'])
                time, longitude = pytplot.get_data(dataset.links['lon'])
                latitude = latitude.transpose()[0]
                longitude = longitude.transpose()[0]
                nearest_time_index = np.abs(time - test_time).argmin()
                lat_point = latitude[nearest_time_index]
                lon_point = longitude[nearest_time_index]
                self.plotwindow.scatterPlot([lon_point], [lat_point], size=pointsize, pen=pg.mkPen(None), brush=color)

        return

    def _setbackground(self):
        if 'alpha' in pytplot.data_quants[self.tvar_name].extras:
            alpha = pytplot.data_quants[self.tvar_name].extras['alpha']
        else:
            alpha = 1
        if 'basemap' in pytplot.data_quants[self.tvar_name].extras:
            if os.path.isfile(pytplot.data_quants[self.tvar_name].extras['basemap']):
                from scipy import misc
                img = misc.imread(pytplot.data_quants[self.tvar_name].extras['basemap'], mode='RGBA')
                # Need to flip the image upside down...This will probably be fixed in
                # a future release, so this will need to be deleted at some point
                img = img[::-1]
                bm = pg.ImageItem(image=img, opacity=alpha)
                bm.setRect(QtCore.QRect(0, -90, 360, 180))
                self.plotwindow.addItem(bm)



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

from __future__ import division
import sys
import os
import pytplot
from bokeh.io import output_file, show, output_notebook, save
from . import HTMLPlotter
from bokeh.embed import components
from pytplot import tplot_utilities
import tempfile

if pytplot.using_graphics:
    from .QtPlotter import PyTPlot_Exporter
    from pyqtgraph.Qt import QtCore, QtGui, QtCore
    import pyqtgraph as pg
    from . import QtPlotter, interactiveplot, staticplot, staticplot_tavg

    try:
        from PyQt5.QtWebKitWidgets import QWebView as WebView
    except:
        from PyQt5.QtWebEngineWidgets import QWebEngineView as WebView


def tplot(name,
          var_label=None,
          auto_color=True,
          interactive=False,
          combine_axes=True,
          nb=False,
          save_file=None,
          gui=False,
          qt=False,
          bokeh=False,
          save_png=None,
          display=True,
          testing=False):
    """
    This is the function used to display the tplot variables stored in memory.
    The default output is to show the plots stacked on top of one another inside a GUI window.
    The GUI window has the option to export the plots in either PNG or HTML formats.

    .. note::
        This plotting routine uses the python Bokeh library, which creates plots using HTML and Javascript.
        Bokeh is technically still in beta, so future patches to Bokeh may require updates to this function.

    Parameters:
        name : str / list
            List of tplot variables that will be plotted
        var_label : str, optional
            The name of the tplot variable you would like as
            a second x axis.
        auto_color : bool, optional
            Automatically color the plot lines.
        interactive : bool, optional
            If True, a secondary interactive plot will be generated next to spectrogram plots.
            Mousing over the spectrogram will display a slice of data from that time on the
            interactive chart.
        combine_axes : bool, optional
            If True, the axes are combined so that they all display the same x range.  This also enables
            scrolling/zooming/panning on one plot to affect all of the other plots simultaneously.
        nb : bool, optional
            If True, the plot will be displayed inside of a current Jupyter notebook session.
        save_file : str, optional
            A full file name and path.
            If this option is set, the plot will be automatically saved to the file name provided in an HTML format.
            The plots can then be opened and viewed on any browser without any requirements.
        bokeh : bool, optional
            If True, plots data using bokeh
            Else (bokeh=False or omitted), plots data using PyQtGraph
        gui : bool, optional
            If True, then this function will output the 2 HTML components of the generated plots as string variables.
            This is useful if you are embedded the plots in your own GUI.  For more information, see
            http://bokeh.pydata.org/en/latest/docs/user_guide/embed.html
        qt : bool, optional
            If True, then this function will display the plot inside of the Qt window.  From this window, you
            can choose to export the plots as either an HTML file, or as a PNG.
        save_png : str, optional
            A full file name and path.
            If this option is set, the plot will be automatically saved to the file name provided in a PNG format.
        display: bool, optional
            If True, then this function will display the plotted tplot variables. Necessary to make this optional
            so we can avoid it in a headless server environment.
        testing: bool, optional
            If True, doesn't run the '(hasattr(sys, 'ps1'))' line that makes plots interactive - i.e., avoiding issues

    Returns:
        None

    Examples:
        >>> #Plot a single line in bokeh
        >>> import pytplot
        >>> x_data = [2,3,4,5,6]
        >>> y_data = [1,2,3,4,5]
        >>> pytplot.store_data("Variable1", data={'x':x_data, 'y':y_data})
        >>> pytplot.tplot("Variable1",bokeh=True)

        >>> #Display two plots
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [[1,5],[2,4],[3,3],[4,2],[5,1]]
        >>> pytplot.store_data("Variable2", data={'x':x_data, 'y':y_data})
        >>> pytplot.tplot(["Variable1", "Variable2"])

        >>> #Display 2 plots, using Variable1 as another x axis
        >>> x_data = [1,2,3]
        >>> y_data = [ [1,2,3] , [4,5,6], [7,8,9] ]
        >>> v_data = [1,2,3]
        >>> pytplot.store_data("Variable3", data={'x':x_data, 'y':y_data, 'v':v_data})
        >>> pytplot.options("Variable3", 'spec', 1)
        >>> pytplot.tplot(["Variable2", "Variable3"], var_label='Variable1')

        >>> #Plot all 3 tplot variables, sending the output to an HTML file
        >>> pytplot.tplot(["Variable1", "Variable2", "Variable3"], save_file='C:/temp/pytplot_example.html')

        >>> #Plot all 3 tplot variables, sending the HTML output to a pair of strings
        >>> div, component = pytplot.tplot(["Variable1", "Variable2", "Variable3"], gui=True)
    """

    if not pytplot.using_graphics and save_file is None:
        print("Qt was not successfully imported.  Specify save_file to save the file as a .html file.")
        return
    # Check a bunch of things
    if not isinstance(name, list):
        name = [name]
        num_plots = 1
    else:
        num_plots = len(name)

    for i in range(num_plots):
        if isinstance(name[i], int):
            name[i] = list(pytplot.data_quants.keys())[name[i]]
        if name[i] not in pytplot.data_quants.keys():
            print(str(i) + " is currently not in pytplot")
            return

    if isinstance(var_label, int):
        var_label = list(pytplot.data_quants.keys())[var_label]

    if bokeh:
        layout = HTMLPlotter.generate_stack(name, var_label=var_label, auto_color=auto_color, combine_axes=combine_axes,
                                            interactive=interactive)
        # Output types
        if gui:
            script, div = components(layout)
            return script, div
        elif nb:
            output_notebook()
            show(layout)
            return
        elif save_file is not None:
            output_file(save_file, mode='inline')
            save(layout)
            return
        elif qt:
            available_qt_window = tplot_utilities.get_available_qt_window()
            dir_path = tempfile.gettempdir()  # send to user's temp directory
            output_file(os.path.join(dir_path, "temp.html"), mode='inline')
            save(layout)
            new_layout = WebView()
            available_qt_window.resize(pytplot.tplot_opt_glob['window_size'][0] + 100,
                                       pytplot.tplot_opt_glob['window_size'][1] + 100)
            new_layout.resize(pytplot.tplot_opt_glob['window_size'][0], pytplot.tplot_opt_glob['window_size'][1])
            dir_path = tempfile.gettempdir()  # send to user's temp directory
            new_layout.setUrl(QtCore.QUrl.fromLocalFile(os.path.join(dir_path, "temp.html")))
            available_qt_window.newlayout(new_layout)
            available_qt_window.show()
            available_qt_window.activateWindow()
            if testing:
                return
            if not (hasattr(sys, 'ps1')) or not hasattr(QtCore, 'PYQT_VERSION'):
                QtGui.QApplication.instance().exec_()
            return
        else:
            dir_path = tempfile.gettempdir()  # send to user's temp directory
            output_file(os.path.join(dir_path, "temp.html"), mode='inline')
            show(layout)
            return
    else:
        if save_png is not None:
            layout = QtPlotter.generate_stack(name, var_label=var_label, combine_axes=combine_axes,
                                              mouse_moved_event=pytplot.hover_time.change_hover_time)
            layout.resize(pytplot.tplot_opt_glob['window_size'][0], pytplot.tplot_opt_glob['window_size'][1])
            for i, item in enumerate(layout.items()):
                if type(item) == pg.graphicsItems.GraphicsLayout.GraphicsLayout:
                    layout.items()[i].resize(pytplot.tplot_opt_glob['window_size'][0],
                                             pytplot.tplot_opt_glob['window_size'][1])
            exporter = PyTPlot_Exporter.PytplotExporter(layout)
            exporter.parameters()['width'] = pytplot.tplot_opt_glob['window_size'][0]
            exporter.parameters()['height'] = pytplot.tplot_opt_glob['window_size'][1]
            exporter.export(save_png)

        if display:
            # Set up all things needed for when a user asks to save plot from window
            layout_orig = QtPlotter.generate_stack(name, var_label=var_label, combine_axes=combine_axes,
                                                   mouse_moved_event=pytplot.hover_time.change_hover_time)
            layout_orig.resize(pytplot.tplot_opt_glob['window_size'][0], pytplot.tplot_opt_glob['window_size'][1])
            for i, item in enumerate(layout_orig.items()):
                if type(item) == pg.graphicsItems.GraphicsLayout.GraphicsLayout:
                    layout_orig.items()[i].resize(pytplot.tplot_opt_glob['window_size'][0],
                                                  pytplot.tplot_opt_glob['window_size'][1])
            exporter = QtPlotter.PytplotExporter(layout_orig)

            # Set up displayed plot window and grab plots to plot on it
            available_qt_window = tplot_utilities.get_available_qt_window()

            layout = QtPlotter.generate_stack(name, var_label=var_label, combine_axes=combine_axes,
                                              mouse_moved_event=pytplot.hover_time.change_hover_time)

            available_qt_window.newlayout(layout)
            available_qt_window.resize(pytplot.tplot_opt_glob['window_size'][0],
                                       pytplot.tplot_opt_glob['window_size'][1])

            # Implement button that lets you save the PNG
            available_qt_window.init_savepng(exporter)

            # Show the plot window and plot
            available_qt_window.show()
            available_qt_window.activateWindow()

            if interactive:
                # Call 2D interactive window; This will only plot something when spectrograms are involved.
                interactiveplot.interactiveplot()

            static_list = [i for i in name if 'static' in pytplot.data_quants[i].extras]
            for tplot_var in static_list:
                # Call 2D static window; This will only plot something when spectrograms are involved.
                staticplot.static2dplot(tplot_var, pytplot.data_quants[tplot_var].extras['static'])

            static_tavg_list = [i for i in name if 'static_tavg' in pytplot.data_quants[i].extras]
            for tplot_var in static_tavg_list:
                # Call 2D static window for time-averaged values; This will only plot something when spectrograms
                # are involved
                staticplot_tavg.static2dplot_timeaveraged(
                    tplot_var, pytplot.data_quants[tplot_var].extras['static_tavg'])

            # (hasattr(sys, 'ps1')) checks to see if we're in ipython
            # plots the plots!
            if testing:
                return
            if not (hasattr(sys, 'ps1')) or not hasattr(QtCore, 'PYQT_VERSION'):
                QtGui.QApplication.instance().exec_()

        return



import pyqtgraph as pg
import numpy as np
import pytplot
from pytplot import tplot_utilities


def static2dplot(var, time):
    """ If the static option is set in tplot, and is supplied with a time, then the spectrogram plot(s) for which
    it is set will have another window pop up, with y and z values plotted at the specified time. """

    # Grab names of data loaded in as tplot variables.
    names = list(pytplot.data_quants.keys())
    # Get data we'll actually work with here.
    valid_variables = tplot_utilities.get_data(names)

    # Don't plot anything unless we have spectrograms with which to work.
    if valid_variables:
        # Get z label
        labels = tplot_utilities.get_labels_axis_types(names)

        # Put together data in easy-to-access format for plots.
        data = {}
        for name in valid_variables:
            bins = tplot_utilities.get_bins(name)
            time_values, z_values = tplot_utilities.get_z_t_values(name)
            data[name] = [bins, z_values, time_values]

        # Set up the 2D static plot
        pytplot.static_window = pg.GraphicsWindow()
        pytplot.static_window.resize(1000, 600)
        pytplot.static_window.setWindowTitle('Static Window')
        plot = pytplot.static_window.addPlot(title='2D Static Plot', row=0, col=0)
        # Make it so that whenever this first starts up, you just have an empty plot
        plot_data = plot.plot([], [])

        if var in valid_variables:
            # Get min/max values of data's time range (in both datetime and seconds since epoch)
            t_min = np.nanmin(time_values)
            t_min_str = tplot_utilities.int_to_str(np.nanmin(time_values))
            t_min_conv_back = tplot_utilities.str_to_int(t_min_str)
            t_max = np.nanmax(time_values)
            t_max_str = tplot_utilities.int_to_str(np.nanmax(time_values))
            t_max_conv_back = tplot_utilities.str_to_int(t_max_str)
            # Convert user input to seconds since epoch
            user_time = tplot_utilities.str_to_int(time)

            # Covering situation where user entered a time not in the dataset!
            # As long as they used a time in the dataset, this will not trigger.
            if user_time not in range(t_min_conv_back, t_max_conv_back+1):
                while True:
                    try:
                        user_time = tplot_utilities.str_to_int(input(
                            'Chosen time not in range of data [{} to {}]. Input new time (%Y-%m-%d %H:%M:%S). '.format(
                                t_min_str, t_max_str)))
                    except:
                        continue
                    else:
                        if user_time not in range(int(t_min), int(t_max)):
                            continue
                        else:
                            break

            # Get time closest to the user's time choice
            time_array = np.array(data[var][2])
            array = np.asarray(time_array)
            idx = (np.abs(array - user_time)).argmin()
            # If user indicated they wanted the interactive plot's axes to be logged, log 'em.
            # But first make sure that values in x and y are loggable!
            x_axis = False
            y_axis = False
            # Checking x axis
            if np.nanmin(data[name][0][:]) < 0:
                print('Negative data is incompatible with log plotting.')
            elif np.nanmin(data[name][0][:]) >= 0 and labels[name][2] == 'log':
                x_axis = True
            # Checking y axis
            if np.nanmin(list(data[name][1][idx])) < 0:
                print('Negative data is incompatible with log plotting')
            elif np.nanmin(list(data[name][1][idx])) >= 0 and labels[name][3] == 'log':
                y_axis = True

            # Set plot labels
            plot.setLabel('bottom', '{}'.format(labels[name][0]))
            plot.setLabel('left', '{}'.format(labels[name][1]))
            plot.setLogMode(x=x_axis, y=y_axis)
            # Update x and y range if user modified it
            tplot_utilities.set_x_range(name, x_axis, plot)
            tplot_utilities.set_y_range(name, y_axis, plot)
            # Plot data based on time we're hovering over
            plot_data.setData(data[name][0][:], list(data[name][1][idx]))



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyqtgraph as pg
from .CustomAxis.NonLinearAxis import NonLinearAxis
from .CustomViewBox.CustomVB import CustomVB
import pytplot


class TVarFigureAxisOnly(pg.GraphicsLayout):
    def __init__(self, tvar_name):
        self.tvar_name = tvar_name
        
        # Sets up the layout of the Tplot Object
        pg.GraphicsLayout.__init__(self)
        self.layout.setHorizontalSpacing(50)
        self.layout.setContentsMargins(0, 0, 0, 0)

        self.labelStyle = {'font-size': str(pytplot.data_quants[self.tvar_name].extras['char_size'])+'pt'}
        
        vb = CustomVB(enableMouse=False)
        yaxis = pg.AxisItem("left")
        yaxis.setLabel(pytplot.data_quants[self.tvar_name].yaxis_opt['axis_label'], **self.labelStyle)
        yaxis.setWidth(100)
        yaxis.label.rotate(90)
        yaxis.label.translate(0, -40)
        xaxis = NonLinearAxis(orientation='bottom', data=pytplot.data_quants[self.tvar_name])
        self.plotwindow = self.addPlot(row=0, col=0, axisItems={'bottom': xaxis, 'left': yaxis}, viewBox=vb, colspan=1)
        self.plotwindow.buttonsHidden = True
        self.plotwindow.setMaximumHeight(20)
        
        # Set up the view box needed for the legends
        self.legendvb = pg.ViewBox(enableMouse=False)
        self.legendvb.setMaximumWidth(100)
        self.addItem(self.legendvb, 0, 1)



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np
from scipy import interpolate
from scipy.interpolate import interp1d

"""
    Linearly interpolates data to user-specified values 
    
    Parameters:
        tvar1 : str
            Name of tvar whose data will be interpolated to specified times.  
        times : int/list
            Desired times for interpolation.
        newtvar : str
            Name of new tvar in which to store interpolated data.
            
    Returns:
        None
    
    Examples:
        >>> Interpolate data for 'd' to values [3,4,5,6,7,18].
        >>> pytplot.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1],[2,2],[100,100],[4,4],[5,5],[6,6],[7,7]]})
        >>> pytplot.tplot_resample('d',[3,4,5,6,7,18],'d_resampled')
    """

def pdresample(tvar, rule, how=None, axis=0, fill_method=None, closed=None, label=None, 
               convention='start', kind=None, loffset=None, limit=None, base=0, on=None, 
               level=None):
    tvar_df = pytplot.data_quants[tvar].data
    pytplot.data_quants[tvar].data = tvar_df.resample(rule, how, axis, fill_method, closed, label, convention, kind, loffset, limit,
                                      base, on, level)
    
    return
                              
def resample(tvar1,times,newtvar='tvar_resample'):
    #create dummy dataframe for times to interpolate to
    pytplot.store_data('times_for_resample',data={'x':times,'y':np.zeros(len(times))})
    df_index = pytplot.data_quants[tvar1].data.columns.copy()
    new_df = []
    spec_df = []
    tvar_orig = pytplot.data_quants[tvar1].data.copy()
     
    if (pytplot.data_quants[tvar1].spec_bins is not None) and (pytplot.data_quants[tvar1].spec_bins_time_varying == True):
        spec_orig = pytplot.data_quants[tvar1].spec_bins.copy()
         
    #for each column of dataframe
    for i in df_index:
        tv2_col = [item[i] for item in tvar_orig.values]
        if (pytplot.data_quants[tvar1].spec_bins is not None) and (pytplot.data_quants[tvar1].spec_bins_time_varying == True):
            spec_col = [item[i] for item in spec_orig.values]
        #linear interpolation
        f = interp1d(tvar_orig.index,tv2_col,fill_value="extrapolate")
        new_df = new_df + [f(times)]
         
        if (pytplot.data_quants[tvar1].spec_bins is not None) and (pytplot.data_quants[tvar1].spec_bins_time_varying == True):
            g = interp1d(tvar_orig.index,spec_col,fill_value="extrapolate")
            spec_df = spec_df + [g(times)]
    new_df = np.transpose((list(new_df)))
     
    if (pytplot.data_quants[tvar1].spec_bins is not None) and (pytplot.data_quants[tvar1].spec_bins_time_varying == True):
        spec_df = np.transpose((list(spec_df)))
    #store interpolated tvar
    if (pytplot.data_quants[tvar1].spec_bins is not None) and (pytplot.data_quants[tvar1].spec_bins_time_varying == True):
        pytplot.store_data(newtvar, data={'x':times,'y':new_df,'v':spec_df})
    else:
        pytplot.store_data(newtvar, data={'x':times,'y':new_df})
    #delete dummy dataframe from pytplot.data_quants
    pytplot.del_data(name='times_for_resample')
    return



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot

#LINEAR INTERPOLATION
#interpolate over NaN data
def interp_nan(tvar1,newtvar='tvar_interp_nan',s_limit=0):
    tv1 = pytplot.data_quants[tvar1].data.copy()
    tv1 = tv1.astype(float)
    cadence = tv1.index[1] - tv1.index[0]
    n_nans = int(round(s_limit/cadence))
    if s_limit == 0:
        tv1 = tv1.interpolate(method='linear')
    else:
        tv1 = tv1.interpolate(method='linear',limit=n_nans,limit_direction='both') 
    tv1 = tv1.astype(object)
    pytplot.store_data(newtvar,data={'x':tv1.index,'y':tv1})
    return tv1


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

from pytplot import data_quants

def link(names, link_name, link_type='alt'):
    
    link_type = link_type.lower()
    if not isinstance(names, list):
        names = [names]
        
    for i in names:
        if i not in data_quants.keys():
            print(str(i) + " is not currently in pytplot.")
            return
        
        if isinstance(i,int):
            i = list(data_quants.keys())[i-1]
    
        data_quants[i].link_to_tvar(link_type, link_name)
                
    return
        
    


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyqtgraph as pg
import numpy as np
import pytplot
from .CustomLegend.CustomLegend import CustomLegendItem
from .CustomAxis.AxisItem import AxisItem
from .CustomViewBox.NoPaddingPlot import NoPaddingPlot


class TVarFigureAlt(pg.GraphicsLayout):
    def __init__(self, tvar_name, show_xaxis=False, mouse_function=None):

        self.tvar_name = tvar_name
        self.show_xaxis = show_xaxis
        self.crosshair = pytplot.tplot_opt_glob['crosshair']

        # Sets up the layout of the Tplot Object
        pg.GraphicsLayout.__init__(self)
        self.layout.setHorizontalSpacing(50)
        self.layout.setContentsMargins(0, 0, 0, 0)
        # Set up the x axis
        self.xaxis = pg.AxisItem(orientation='bottom')
        self.xaxis.setHeight(35)
        self.xaxis.enableAutoSIPrefix(enable=False)
        # Set up the y axis
        self.yaxis = AxisItem("left")
        self.yaxis.setWidth(100)

        vb = NoPaddingPlot()
        self.plotwindow = self.addPlot(row=0, col=0, axisItems={'bottom': self.xaxis, 'left': self.yaxis}, viewBox=vb)

        # Set up the view box needed for the legends
        self.legendvb = pg.ViewBox(enableMouse=False)
        self.legendvb.setMaximumWidth(100)
        self.legendvb.setXRange(0, 1, padding=0)
        self.legendvb.setYRange(0, 1, padding=0)
        self.addItem(self.legendvb, 0, 1)

        self.curves = []
        self.colors = self._setcolors()
        self.colormap = self._setcolormap()

        self.labelStyle = {'font-size': str(pytplot.data_quants[self.tvar_name].extras['char_size'])+'pt'}

        if show_xaxis:
            self.plotwindow.showAxis('bottom')
        else:
            self.plotwindow.hideAxis('bottom')

        self._mouseMovedFunction = mouse_function

        self.vLine = pg.InfiniteLine(angle=90, movable=False, pen=pg.mkPen('k'))
        self.hLine = pg.InfiniteLine(angle=0, movable=False, pen=pg.mkPen('k'))
        self.plotwindow.addItem(self.vLine, ignoreBounds=True)
        self.plotwindow.addItem(self.hLine, ignoreBounds=True)
        self.vLine.setVisible(False)
        self.hLine.setVisible(False)

        self.label = pg.LabelItem(justify='left')
        self.addItem(self.label, row=1, col=0)

        # Set legend options
        self.hoverlegend = CustomLegendItem(offset=(0, 0))
        # Allow the user to set x-axis(time) and y-axis data names in crosshairs
        self.hoverlegend.setItem(pytplot.data_quants[self.tvar_name].xaxis_opt['crosshair'] + ':', "0")
        self.hoverlegend.setItem(pytplot.data_quants[self.tvar_name].yaxis_opt['crosshair'] + ':', "0")

        self.hoverlegend.setVisible(False)
        self.hoverlegend.setParentItem(self.plotwindow.vb)

    def buildfigure(self):
        self._setxrange()
        self._setyrange()
        self._setyaxistype()
        self._setzaxistype()
        self._setzrange()
        self._visdata()
        self._addtimebars()
        self._setxaxislabel()
        self._setyaxislabel()
        if self.crosshair:
            self._addmouseevents()
        self._addlegend()

    def getfig(self):
        return self

    def _setxaxislabel(self):
        self.xaxis.setLabel("Altitude", **self.labelStyle)

    def _setyaxislabel(self):
        self.yaxis.setLabel(pytplot.data_quants[self.tvar_name].yaxis_opt['axis_label'], **self.labelStyle)

    def _setyaxistype(self):
        if self._getyaxistype() == 'log':
            self.plotwindow.setLogMode(y=True)
        else:
            self.plotwindow.setLogMode(y=False)
        return

    def _getyaxistype(self):
        if 'y_axis_type' in pytplot.data_quants[self.tvar_name].yaxis_opt:
            return pytplot.data_quants[self.tvar_name].yaxis_opt['y_axis_type']
        else:
            return 'linear'

    def _setxrange(self):
        if 'alt_range' in pytplot.tplot_opt_glob:
            self.plotwindow.setXRange(pytplot.tplot_opt_glob['alt_range'][0], pytplot.tplot_opt_glob['alt_range'][1])
        else:
            return

    def getaxistype(self):
        axis_type = 'altitude'
        link_y_axis = False
        return axis_type, link_y_axis

    def _addmouseevents(self):
        if self.plotwindow.scene() is not None:
            self.plotwindow.scene().sigMouseMoved.connect(self._mousemoved)

    def _mousemoved(self, evt):
        # get current position
        pos = evt
        # if plot window contains position
        if self.plotwindow.sceneBoundingRect().contains(pos):
            mousepoint = self.plotwindow.vb.mapSceneToView(pos)
            # grab x and y mouse locations
            index_x = int(mousepoint.x())
            index_y = int(mousepoint.y())
            # add crosshairs
            if self._mouseMovedFunction is not None:
                self._mouseMovedFunction(int(mousepoint.x()))
                self.vLine.setPos(mousepoint.x())
                self.hLine.setPos(mousepoint.y())
                self.vLine.setVisible(True)
                self.hLine.setVisible(True)

            # Set legend options
            self.hoverlegend.setVisible(True)
            # Allow the user to set x-axis(time) and y-axis data names in crosshairs
            self.hoverlegend.setItem(pytplot.data_quants[self.tvar_name].xaxis_opt['crosshair'] + ':', index_x)
            self.hoverlegend.setItem(pytplot.data_quants[self.tvar_name].yaxis_opt['crosshair'] + ':', index_y)
        else:
            self.hoverlegend.setVisible(False)
            self.vLine.setVisible(False)
            self.hLine.setVisible(False)

    def _addlegend(self):
        if 'legend_names' in pytplot.data_quants[self.tvar_name].yaxis_opt:
            legend_names = pytplot.data_quants[self.tvar_name].yaxis_opt['legend_names']
            if len(legend_names) != len(self.curves):
                print("Number of lines do not match length of legend names")
            if len(legend_names) == 1:
                pos_array = [.5]
            else:
                pos_array = np.linspace(1, 0, len(legend_names))
            i = 0
            for legend_name in legend_names:
                if i + 1 == len(legend_names):  # Last
                    text = pg.TextItem(text=legend_name, anchor=(0, 1.5), color=self.colors[i % len(self.colors)])
                elif i == 0:  # First
                    text = pg.TextItem(text=legend_name, anchor=(0, -.5), color=self.colors[i % len(self.colors)])
                else:  # All others
                    text = pg.TextItem(text=legend_name, anchor=(0, 0.5), color=self.colors[i % len(self.colors)])
                self.legendvb.addItem(text)
                text.setPos(0, pos_array[i])
                i += 1

    def _setzaxistype(self):
        if self._getzaxistype() == 'log':
            self.zscale = 'log'
        else:
            self.zscale = 'linear'

    def _getzaxistype(self):
        return

    def _setcolors(self):
        if 'line_color' in pytplot.data_quants[self.tvar_name].extras:
            return pytplot.data_quants[self.tvar_name].extras['line_color']
        else:
            return pytplot.tplot_utilities.rgb_color(['k', 'r', 'seagreen', 'b', 'darkturquoise', 'm', 'goldenrod'])

    def _setcolormap(self):
        return

    def _setyrange(self):
        if self._getyaxistype() == 'log':
            if pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][0] < 0 or \
                    pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][1] < 0:
                return
            self.plotwindow.vb.setYRange(np.log10(pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][0]),
                                         np.log10(pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][1]),
                                         padding=0)
        else:
            self.plotwindow.vb.setYRange(pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][0],
                                         pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][1], padding=0)

    def _setzrange(self):
        return

    def _addtimebars(self):
        # initialize dataset variable
        datasets = []
        # grab tbardict
        tbardict = pytplot.data_quants[self.tvar_name].time_bar
        ltbar = len(tbardict)
        # make sure data is in list format
        if isinstance(pytplot.data_quants[self.tvar_name].data, list):
            for oplot_name in pytplot.data_quants[self.tvar_name].data:
                datasets.append(pytplot.data_quants[oplot_name])
        else:
            datasets.append(pytplot.data_quants[self.tvar_name])
        for dataset in datasets:
            # for location in tbar dict
            for i in range(ltbar):
                # get times, color, point size
                test_time = pytplot.data_quants[self.tvar_name].time_bar[i]["location"]
                color = pytplot.data_quants[self.tvar_name].time_bar[i]["line_color"]
                pointsize = pytplot.data_quants[self.tvar_name].time_bar[i]["line_width"]
                # correlate given time with corresponding data/alt points
                time, altitude = pytplot.get_data(dataset.links['alt'])
                altitude = altitude.transpose()[0]
                nearest_time_index = np.abs(time - test_time).argmin()
                data_point = dataset.data.iloc[nearest_time_index][0]
                alt_point = altitude[nearest_time_index]
                self.plotwindow.scatterPlot([alt_point], [data_point], size=pointsize, pen=pg.mkPen(None), brush=color)
        return

    def _visdata(self):
        datasets = []
        if isinstance(pytplot.data_quants[self.tvar_name].data, list):
            for oplot_name in pytplot.data_quants[self.tvar_name].data:
                datasets.append(pytplot.data_quants[oplot_name])
        else:
            datasets.append(pytplot.data_quants[self.tvar_name])
        line_num = 0
        for dataset in datasets:
            for i in range(0, len(dataset.data.columns)):
                t_link, x = pytplot.get_data(dataset.links['alt'])
                # Need to trim down the data points to fit within the link
                t_tvar = dataset.data.index.values
                data = dataset.data[i].values
                while t_tvar[-1] > t_link[-1]:
                    t_tvar = np.delete(t_tvar, -1)
                    data = np.delete(data, -1)
                while t_tvar[0] < t_link[0]:
                    t_tvar = np.delete(t_tvar, 0)
                    data = np.delete(data, 0)

                x = x.transpose()[0]
                self.curves.append(self.plotwindow.scatterPlot(x.tolist(), data.tolist(),
                                                               pen=pg.mkPen(None), brush=self.colors[line_num %
                                                                                                     len(self.colors)]))
                line_num += 1



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyqtgraph as pg
import numpy as np
import pytplot
from .CustomAxis.DateAxis import DateAxis
from .CustomLegend.CustomLegend import CustomLegendItem
from .CustomAxis.AxisItem import AxisItem
from .CustomViewBox.NoPaddingPlot import NoPaddingPlot
from .CustomLinearRegionItem.CustomLinearRegionItem import CustomLinearRegionItem


class TVarFigure1D(pg.GraphicsLayout):
    def __init__(self, tvar_name, show_xaxis=False, mouse_function=None):

        self.tvar_name = tvar_name
        self.show_xaxis = show_xaxis
        self.crosshair = pytplot.tplot_opt_glob['crosshair']

        # Sets up the layout of the Tplot Object
        pg.GraphicsLayout.__init__(self)
        self.layout.setHorizontalSpacing(50)
        self.layout.setContentsMargins(0, 0, 0, 0)
        # Set up the x axis
        self.xaxis = DateAxis(orientation='bottom')
        self.xaxis.setHeight(35)
        self.xaxis.enableAutoSIPrefix(enable=False)
        # Set up the y axis
        self.yaxis = AxisItem("left")
        self.yaxis.setWidth(100)

        vb = NoPaddingPlot()
        self.plotwindow = self.addPlot(row=0, col=0, axisItems={'bottom': self.xaxis, 'left': self.yaxis}, viewBox=vb)

        # Set up the view box needed for the legends
        self.legendvb = pg.ViewBox(enableMouse=False)
        self.legendvb.setMaximumWidth(100)
        self.legendvb.setXRange(0, 1, padding=0)
        self.legendvb.setYRange(0, 1, padding=0)
        self.addItem(self.legendvb, 0, 1)

        self.curves = []
        self.colors = self._setcolors()
        self.colormap = self._setcolormap()

        self.labelStyle = {'font-size': str(pytplot.data_quants[self.tvar_name].extras['char_size'])+'pt'}

        if show_xaxis:
            self.plotwindow.showAxis('bottom')
        else:
            self.plotwindow.hideAxis('bottom')

        self._mouseMovedFunction = mouse_function

        self.label = pg.LabelItem(justify='left')
        self.addItem(self.label, row=1, col=0)

        # Set legend options
        self.hoverlegend = CustomLegendItem(offset=(0, 0))
        self.hoverlegend.setItem("Date:", "0")
        # Allow the user to set x-axis(time) and y-axis names in crosshairs
        self.hoverlegend.setItem(pytplot.data_quants[self.tvar_name].xaxis_opt['crosshair'] + ':', "0")
        self.hoverlegend.setItem(pytplot.data_quants[self.tvar_name].yaxis_opt['crosshair'] + ':', "0")
        self.hoverlegend.setVisible(False)
        self.hoverlegend.setParentItem(self.plotwindow.vb)

    def _set_crosshairs(self):
        self.vLine = pg.InfiniteLine(angle=90, movable=False, pen=pg.mkPen('k'))
        self.hLine = pg.InfiniteLine(angle=0, movable=False, pen=pg.mkPen('k'))
        self.plotwindow.addItem(self.vLine, ignoreBounds=True)
        self.plotwindow.addItem(self.hLine, ignoreBounds=True)
        self.vLine.setVisible(False)
        self.hLine.setVisible(False)

    def _set_roi_lines(self, dataset):
        # Locating the two times between which there's a roi
        time = dataset.data.index.tolist()
        roi_1 = pytplot.tplot_utilities.str_to_int(pytplot.tplot_opt_glob['roi_lines'][0][0])
        roi_2 = pytplot.tplot_utilities.str_to_int(pytplot.tplot_opt_glob['roi_lines'][0][1])
        # find closest time to user-requested time
        x = np.asarray(time)
        x_sub_1 = abs(x - roi_1 * np.ones(len(x)))
        x_sub_2 = abs(x - roi_2 * np.ones(len(x)))
        x_argmin_1 = np.nanargmin(x_sub_1)
        x_argmin_2 = np.nanargmin(x_sub_2)
        x_closest_1 = x[x_argmin_1]
        x_closest_2 = x[x_argmin_2]
        # Create roi box
        roi = CustomLinearRegionItem(orientation=pg.LinearRegionItem.Vertical, values=[x_closest_1, x_closest_2])
        roi.setBrush([211, 211, 211, 130])
        roi.lines[0].setPen('r', width=2.5)
        roi.lines[1].setPen('r', width=2.5)
        self.plotwindow.addItem(roi)

    def buildfigure(self):
        self._setxrange()
        self._setyrange()
        self._setyaxistype()
        self._setzaxistype()
        self._setzrange()
        self._visdata()
        self._setxaxislabel()
        self._setyaxislabel()
        self._addlegend()
        self._addtimebars()
        if self.crosshair:
            self._set_crosshairs()
            self._addmouseevents()

    def _setxaxislabel(self):
        self.xaxis.setLabel(pytplot.data_quants[self.tvar_name].xaxis_opt['axis_label'], **self.labelStyle)

    def _setyaxislabel(self):
        self.yaxis.setLabel(pytplot.data_quants[self.tvar_name].yaxis_opt['axis_label'], **self.labelStyle)

    def getfig(self):
        return self

    def _visdata(self):
        datasets = []
        if isinstance(pytplot.data_quants[self.tvar_name].data, list):
            for oplot_name in pytplot.data_quants[self.tvar_name].data:
                datasets.append(pytplot.data_quants[oplot_name])
        else:
            datasets.append(pytplot.data_quants[self.tvar_name])
        line_num = 0
        for dataset in datasets:
            # Add region of interest (roi) lines if applicable
            if 'roi_lines' in pytplot.tplot_opt_glob.keys():
                self._set_roi_lines(dataset)
            for i in range(0, len(dataset.data.columns)):
                self.curves.append(self.plotwindow.plot(dataset.data.index.tolist(),
                                                        dataset.data[i].tolist(),
                                                        pen=self.colors[line_num % len(self.colors)]))
                line_num += 1

    def _setyaxistype(self):
        if self._getyaxistype() == 'log':
            self.plotwindow.setLogMode(y=True)
        else:
            self.plotwindow.setLogMode(y=False)
        return

    def _addlegend(self):
        if 'legend_names' in pytplot.data_quants[self.tvar_name].yaxis_opt:
            legend_names = pytplot.data_quants[self.tvar_name].yaxis_opt['legend_names']
            n_items = len(legend_names)
            bottom_bound = 0.5 + (n_items - 1) * 0.05
            top_bound = 0.5 - (n_items - 1) * 0.05
            if len(legend_names) != len(self.curves):
                print("Number of lines do not match length of legend names")
            if len(legend_names) == 1:
                pos_array = [.5]
            else:
                pos_array = np.linspace(bottom_bound, top_bound, len(legend_names))
            i = 0
            for legend_name in legend_names:
                def rgb(red, green, blue): return '#%02x%02x%02x' % (red, green, blue)
                r = self.colors[i % len(self.colors)][0]
                g = self.colors[i % len(self.colors)][1]
                b = self.colors[i % len(self.colors)][2]
                hex_num = rgb(r, g, b)
                color_text = 'color: ' + hex_num
                font_size = 'font-size: '+str(pytplot.data_quants[self.tvar_name].extras['char_size'])+'pt'
                opts = [color_text, font_size]
                full = "<span style='%s'>%s</span>" % ('; '.join(opts), legend_name)
                if i + 1 == len(legend_names):  # Last
                    text = pg.TextItem(anchor=(0, 0.5))
                    text.setHtml(full)
                elif i == 0:  # First
                    text = pg.TextItem(anchor=(0, 0.5))
                    text.setHtml(full)
                else:  # All others
                    text = pg.TextItem(anchor=(0, 0.5))
                    text.setHtml(full)
                self.legendvb.addItem(text)
                text.setPos(0, pos_array[i])
                i += 1

    def _addmouseevents(self):
        if self.plotwindow.scene() is not None:
            self.plotwindow.scene().sigMouseMoved.connect(self._mousemoved)

    def _mousemoved(self, evt):
        # get current position
        pos = evt
        # if plot window contains position
        if self.plotwindow.sceneBoundingRect().contains(pos):
            mousepoint = self.plotwindow.vb.mapSceneToView(pos)
            # grab x and y mouse locations
            index_x = int(mousepoint.x())
            index_y = round(float(mousepoint.y()), 4)
            date = (pytplot.tplot_utilities.int_to_str(index_x))[0:10]
            time = (pytplot.tplot_utilities.int_to_str(index_x))[11:19]
            # add crosshairs
            if self._mouseMovedFunction is not None:
                self._mouseMovedFunction(int(mousepoint.x()))
                self.vLine.setPos(mousepoint.x())
                self.hLine.setPos(mousepoint.y())
                self.vLine.setVisible(True)
                self.hLine.setVisible(True)

            self.hoverlegend.setVisible(True)
            self.hoverlegend.setItem("Date:", date)
            # Allow the user to set x-axis(time) and y-axis data names in crosshairs
            self.hoverlegend.setItem(pytplot.data_quants[self.tvar_name].xaxis_opt['crosshair'] + ':', time)
            self.hoverlegend.setItem(pytplot.data_quants[self.tvar_name].yaxis_opt['crosshair'] + ':', str(index_y))

        else:
            self.hoverlegend.setVisible(False)
            self.vLine.setVisible(False)
            self.hLine.setVisible(False)

    def _getyaxistype(self):
        if 'y_axis_type' in pytplot.data_quants[self.tvar_name].yaxis_opt:
            return pytplot.data_quants[self.tvar_name].yaxis_opt['y_axis_type']
        else:
            return 'linear'

    def _setzaxistype(self):
        if self._getzaxistype() == 'log':
            self.zscale = 'log'
        else:
            self.zscale = 'linear'

    def _getzaxistype(self):
        return

    def _setcolors(self):
        if 'line_color' in pytplot.data_quants[self.tvar_name].extras:
            return pytplot.data_quants[self.tvar_name].extras['line_color']
        else:
            return pytplot.tplot_utilities.rgb_color(['k', 'r', 'seagreen', 'b', 'darkturquoise', 'm', 'goldenrod'])

    def _setcolormap(self):
        return

    def getaxistype(self):
        axis_type = 'time'
        link_y_axis = False
        return axis_type, link_y_axis

    def _setxrange(self):
        # Check if x range is set.  Otherwise, x range is automatic
        if 'x_range' in pytplot.tplot_opt_glob:
            self.plotwindow.setXRange(pytplot.tplot_opt_glob['x_range'][0], pytplot.tplot_opt_glob['x_range'][1])

    def _setyrange(self):
        if self._getyaxistype() == 'log':
            if pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][0] < 0 or \
                    pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][1] < 0:
                return
            self.plotwindow.vb.setYRange(np.log10(pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][0]),
                                         np.log10(pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][1]),
                                         padding=0)
        else:
            self.plotwindow.vb.setYRange(pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][0],
                                         pytplot.data_quants[self.tvar_name].yaxis_opt['y_range'][1], padding=0)

    def _setzrange(self):
        return

    def _addtimebars(self):
        # find number of times to plot
        dict_length = len(pytplot.data_quants[self.tvar_name].time_bar)
        # for each time
        for i in range(dict_length):
            # pull date, color, thickness
            date_to_highlight = pytplot.data_quants[self.tvar_name].time_bar[i]["location"]
            color = pytplot.data_quants[self.tvar_name].time_bar[i]["line_color"]
            thick = pytplot.data_quants[self.tvar_name].time_bar[i]["line_width"]
            # make infinite line w/ parameters
            infline = pg.InfiniteLine(pos=date_to_highlight, pen=pg.mkPen(color, width=thick))
            # add to plot window
            self.plotwindow.addItem(infline)

        return



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot

#ADD ACROSS COLUMNS
#add tvar data across columns, store in new_tvar
def add_across(tvar1,new_tvar='tvar_aa'):
    #separate and add data
    time = pytplot.data_quants[tvar1].data.index.copy()
    data1 = pytplot.data_quants[tvar1].data.copy()
    data = data1.sum(axis=1)
    #store added data
    pytplot.store_data(new_tvar,data={'x':time, 'y':data})
    return new_tvar



import numpy as np
import datetime as dt
from pyspedas.utilities.time_string import time_string
from pyspedas.utilities.time_double import time_double

def mms_feeps_remove_bad_data(probe = '1', data_rate = 'srvy', datatype = 'electron', level = 'l2', suffix = ''):
    data_rate_level = data_rate + '_' + level

    # electrons first, remove bad eyes
    #;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; 1. BAD EYES ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    #  First, here is a list of the EYES that are bad, we need to make sure these 
    #  data are not usable (i.e., make all of the counts/rate/flux data from these eyes NAN). 
    #  These are for all modes, burst and survey:
bad_data_table = {}
bad_data_table['2017-10-01'] = {}
bad_data_table['2017-10-01']['mms1'] = {'top': [1], 'bottom': [1, 11]}
bad_data_table['2017-10-01']['mms2'] = {'top': [5, 7, 12], 'bottom': [7]}
bad_data_table['2017-10-01']['mms3'] = {'top': [2, 12], 'bottom': [2, 5, 11]}
bad_data_table['2017-10-01']['mms4'] = {'top': [1, 2, 7], 'bottom': [2, 4, 5, 10, 11]}

bad_data_table['2018-10-01'] = {}
bad_data_table['2018-10-01']['mms1'] = {'top': [1], 'bottom': [1, 11]}
bad_data_table['2018-10-01']['mms2'] = {'top': [7, 12], 'bottom': [2, 12]}
bad_data_table['2018-10-01']['mms3'] = {'top': [1, 2], 'bottom': [5, 11]}
bad_data_table['2018-10-01']['mms4'] = {'top': [1, 7], 'bottom': [4, 11]}

dates = np.asarray(time_double(list(bad_data_table.keys())))
closest_table_tm = (np.abs(dates - dt.datetime.now().timestamp())).argmin()

closest_table = time_string(dates[closest_table_tm], '%Y-%m-%d')
bad_data = bad_data_table[closest_table]['mms'+probe]






# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import pandas as pd

#JOIN TVARS
#join TVars into single TVar with multiple columns
def join_vec(tvars,newtvar='tvar_join'):
    df = pytplot.data_quants[tvars[0]].data
    for i,val in enumerate(tvars):
        if i == 0:
            pass
        else:
            df = pd.concat([df,pytplot.data_quants[val].data],axis=1)
    pytplot.store_data(newtvar,data={'x':df.index,'y':df})
    return newtvar


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

from .add_across_partial import add_across_partial
from .add_across import add_across
from .add import add
from .avg_res_data import avg_res_data
from .clip import clip
from .crop import crop
from .deflag import deflag
from .degap import degap
from .derive import derive
from .divide import divide
from .flatten_full import flatten_full
from .flatten import flatten
from .interp_nan import interp_nan
from .interpolate import interpolate
from .join_vec import join_vec
from .multiply import multiply
from .resample import resample, pdresample
from .spec_mult import spec_mult
from .split_vec import split_vec
from .subtract import subtract


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np

#DIVIDE
#divide two tvar data arrays, store in new_tvar
def divide(tvar1,tvar2,new_tvar='tvar_divide',interp='linear'):
    #interpolate tvars
    tv1,tv2 = pytplot.interpolate(tvar1,tvar2,interp=interp)
    #separate and divide data
    time = pytplot.data_quants[tv1].data.index
    data1 = pytplot.data_quants[tv1].data
    data2 = pytplot.data_quants[tv2].data
    data = data1/data2
    #if division by 0, replace with NaN
    data = data.replace([np.inf,-np.inf],np.nan)
    #store divided data
    pytplot.store_data(new_tvar,data={'x':time, 'y':data})
    return new_tvar


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot

#ADD TWO ARRAYS
#add two tvar data arrays, store in new_tvar
def add(tvar1,tvar2,new_tvar=None,interp='linear'):
    #interpolate tvars
    tv1,tv2 = pytplot.interpolate(tvar1,tvar2,interp=interp)
    #separate and add data
    time = pytplot.data_quants[tv1].data.index.copy()
    data1 = pytplot.data_quants[tv1].data.copy()
    data2 = pytplot.data_quants[tv2].data.copy()
    data = data1+data2
    
    if new_tvar == None:
        new_tvar = tvar1 + '_+_' + tvar2
    #store added data
    pytplot.store_data(new_tvar,data={'x':time, 'y':data})
    return new_tvar


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np

#DATA CROPPING
#crop tvar arrays to same timespan
def crop(tvar1,tvar2):
    #grab time and data arrays
    tv1_t = np.asarray(pytplot.data_quants[tvar1].data.index.copy().tolist())
    tv1_d = np.asarray(pytplot.data_quants[tvar1].data.copy())
    tv2_t = np.asarray(pytplot.data_quants[tvar2].data.index.copy().tolist())
    tv2_d = np.asarray(pytplot.data_quants[tvar2].data.copy())
    #find first and last time indices
    t0_1 = tv1_t[0]
    t0_2 = tv2_t[0]
    tx_1 = tv1_t[-1]
    tx_2 = tv2_t[-1]
    #find cut locations
    cut1 = max([t0_1, t0_2])
    cut2 = min([tx_1, tx_2])
    #trim data
    while tv1_t[-1] > cut2:
        tv1_t = np.delete(tv1_t,-1,axis=0)
        tv1_d = np.delete(tv1_d,-1,axis=0)
    while tv1_t[0] < cut1:
        tv1_t = np.delete(tv1_t,0,axis=0)
        tv1_d = np.delete(tv1_d,0,axis=0)
    while tv2_t[-1] > cut2:
        tv2_t = np.delete(tv2_t,-1,axis=0)
        tv2_d = np.delete(tv2_d,-1,axis=0)
    while tv2_t[0] < cut1:
        tv2_t = np.delete(tv2_t,0,axis=0)
        tv2_d = np.delete(tv2_d,0,axis=0)
    #return time and data arrays
    return tv1_t,tv1_d,tv2_t,tv2_d


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot

#PARTIAL FLATTEN
#take average of each column of data, divide column by average over specified time
def flatten(tvar1,start_t,end_t,new_tvar='tvar_flat'):
    df = pytplot.data_quants[tvar1].data
    time = df.index
    #if time given not an index, choose closest time
    if start_t not in time:
        tdiff = abs(time - start_t)
        start_t = time[tdiff.argmin()]
    if end_t not in time:
        tdiff = abs(time - end_t)
        end_t = time[tdiff.argmin()]
    df_index = list(df.columns)
    #divide by specified time average
    for i in df_index:
        df[i] = df[i]/((df.loc[start_t:end_t])[i]).mean()
    pytplot.store_data(new_tvar,data = {'x':df.index,'y':df})
    return new_tvar


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot

#FULL FLATTEN
#take average of each column of data, divide column by column average
def flatten_full(tvar1,new_tvar='tvar_ff'):
    df = pytplot.data_quants[tvar1].data
    df_index = list(df.columns)
    #divide by column average
    for i in df_index:
        df[i] = df[i]/df[i].mean()
    pytplot.store_data(new_tvar,data = {'x':df.index,'y':df})
    return new_tvar


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np

"""
    Change out-of-bounds data to NaN.
    
    Parameters:
        tvar1 : str
            Name of tvar to use for data clipping.  
        ymin : int/float
            Minimum value to keep (inclusive)
        ymax : int/float
            Maximum value to keep (inclusive)
        newtvar : str
            Name of new tvar for clipped data storage
            
    Returns:
        None
    
    Examples:
        >>> Make any values below 2 and above 6 equal to NaN.
        >>> pytplot.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1],[2,2],[100,100],[4,4],[5,5],[6,6],[7,7]]})
        >>> pytplot.clip('d',2,6,'e')
    """

                              
def clip(tvar1,ymin,ymax,newtvar='tvar_clip'):
    #grab column indices
    df_index = pytplot.data_quants[tvar1].data.columns.copy()
    new_df = []
    tvar_orig = pytplot.data_quants[tvar1]
    #for each column of dataframe
    for i in df_index:
        tv2_col = [item[i] for item in tvar_orig.data.values]
        for j,valj in enumerate(tv2_col):
            #if value in column out of specified range, substitute NaN
            if (valj < ymin) or (valj > ymax):
                tv2_col[j] = np.NaN
        new_df = new_df + [tv2_col]
    new_df = np.transpose((list(new_df)))
    #store clipped tvar
    if pytplot.data_quants[tvar1].spec_bins is not None:
        pytplot.store_data(newtvar, data={'x':tvar_orig.data.index,'y':new_df,'v':pytplot.data_quants[tvar1].spec_bins})
    else:
        pytplot.store_data(newtvar, data={'x':tvar_orig.data.index,'y':new_df})
    return



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import numpy as np
from _collections import OrderedDict
from . import HTMLPlotter


# This variable will be constantly changed depending on what x value the user is hovering over
class HoverTime(object):
    hover_time = 0
    functions_to_call = []

    def register_listener(self, fn):
        self.functions_to_call.append(fn)
        return

    # New time is time we're hovering over, grabbed from TVarFigure(1D/Spec/Alt/Map)
    # name is whatever tplot we're currently hovering over (relevant for 2D interactive
    # plots that appear when plotting a spectrogram).
    def change_hover_time(self, new_time, name=None):
        self.hover_time = new_time
        for f in self.functions_to_call:
            f(self.hover_time, name)
        return


using_graphics = True

try:
    import pyqtgraph as pg
    from pyqtgraph.Qt import QtWidgets

    pg.setConfigOptions(imageAxisOrder='row-major')
    pg.setConfigOptions(background='w')


    class PlotWindow(QtWidgets.QMainWindow):
        def __init__(self):
            super().__init__()

        def init_savepng(self, exporter):
            # Set up the save PNG button/call exportpng function that activates when user presses button
            exportdatapngaction = QtWidgets.QAction("Save PNG", self)
            exportdatapngaction.triggered.connect(lambda: self.exportpng(exporter))

            # Set up menu bar to display and call creation of save PNG button
            menubar = self.menuBar()
            menubar.setNativeMenuBar(False)
            menubar.addAction(exportdatapngaction)
            self.setWindowTitle('PyTplot Window')

        def exportpng(self, exporter):
            # Function called by save PNG button to grab the image from the plot window and save it
            fname = QtWidgets.QFileDialog.getSaveFileName(self, 'Open file', 'pytplot.png', filter="png (*.png *.)")
            exporter.parameters()['width'] = tplot_opt_glob['window_size'][0]
            exporter.parameters()['height'] = tplot_opt_glob['window_size'][1]
            exporter.export(fname[0])

        def newlayout(self, layout):
            # Needed for displaying plots
            self.setCentralWidget(layout)
except:
    using_graphics = False


class TVar(object):
    """
    The basic data object in pytplot.  Each dataset is its own separate TVar object.
    This exists to encapsulate the data and details about how to plot the data.
    """

    def __init__(self, name, number, data, spec_bins, xaxis_opt, yaxis_opt, zaxis_opt, line_opt,
                 trange, dtype, create_time, time_bar, extras, links):

        # Name of the TVar
        self.name = name
        # TVar number
        self.number = number
        # The data of the TVar
        self.data = data
        # The spec_bins, if applicable
        self.spec_bins = spec_bins
        # Specifies if extra (interactive) 2D plot appears - only applicable with spectrogram plots
        self.interactive = False
        # If there are spec_bins (i.e., this is a spectrogram), interactive is True
        if self.spec_bins is not None:
            self.interactive = True
        # Dictionary of the x axis options
        self.xaxis_opt = xaxis_opt
        # Dictionary of the y axis options
        self.yaxis_opt = yaxis_opt
        # Dictionary of the z axis options
        self.zaxis_opt = zaxis_opt
        # Dictionary of the x axis options for interactive plots
        self.interactive_xaxis_opt = xaxis_opt
        # Dictionary of the y axis options for interactive plots
        self.interactive_yaxis_opt = yaxis_opt
        # Dictionary of line options
        self.line_opt = line_opt
        # The time range
        self.trange = trange
        # The data type of the data (ex - int/double)
        self.dtype = dtype
        # String of creation time of this object
        self.create_time = create_time
        # Array of time bar objects
        self.time_bar = time_bar
        # Dictionary of extra objects
        self.extras = extras
        # Dictionary of linked tvars (things like latitude/longitude/altitude)
        self.links = links
        # Whether or not the spec_bins vary in time
        self.spec_bins_time_varying = False
        # Whether the spec_bins are ascending or decending order
        self.spec_bins_ascending = self._check_spec_bins_ordering()

        # Add in default values for crosshair names, y/z axis_opt, etc.
        self.extras['char_size'] = 10

        self.xaxis_opt['axis_label'] = 'Time'
        self.yaxis_opt['axis_label'] = 'Y-Axis'
        self.zaxis_opt['axis_label'] = 'Z-Axis'

        self.xaxis_opt['crosshair'] = 'X'
        self.yaxis_opt['crosshair'] = 'Y'
        self.zaxis_opt['crosshair'] = 'Z'

        self.xaxis_opt['x_axis_type'] = 'linear'
        self.yaxis_opt['y_axis_type'] = 'linear'
        self.zaxis_opt['z_axis_type'] = 'linear'

        self.interactive_xaxis_opt['xi_axis_type'] = 'linear'
        self.interactive_yaxis_opt['yi_axis_type'] = 'linear'
        
    def _check_spec_bins_ordering(self):
        """
        This is a private function of the TVar object, this is run during
        object creation to check if spec_bins are ascending or descending
        """
        if self.spec_bins is None:
            return
        if len(self.spec_bins) == len(self.data.index):
            self.spec_bins_time_varying = True
            break_top_loop = False
            for index, row in self.spec_bins.iterrows():
                if row.isnull().values.all():
                    continue
                else:
                    for i in row.index:
                        if np.isfinite(row[i]) and np.isfinite(row[i + 1]):
                            ascending = row[i] < row[i + 1]
                            break_top_loop = True
                            break
                        else:
                            continue
                    if break_top_loop:
                        break
        else:
            ascending = self.spec_bins[0].iloc[0] < self.spec_bins[1].iloc[0]
        return ascending

    def link_to_tvar(self, name, link, method='linear'):
        from scipy import interpolate
        from scipy.interpolate import interp1d
        from .store_data import store_data
        # pull saved variables from data_quants
        link_timeorig = np.asarray(data_quants[link].data.index.tolist())
        link_dataorig = np.asarray(data_quants[link].data[0].tolist())
        tvar_timeorig = np.asarray(self.data.index.tolist())

        # shorten tvar array to be within link array
        while tvar_timeorig[-1] > link_timeorig[-1]:
            tvar_timeorig = np.delete(tvar_timeorig, -1)
        while tvar_timeorig[0] < link_timeorig[0]:
            tvar_timeorig = np.delete(tvar_timeorig, 0)

        x = link_timeorig
        y = link_dataorig
        xnew = tvar_timeorig

        # choose method, interpolate, plot, and store
        if method == 'linear':
            f = interp1d(x, y)
            newvarname = link + "_" + self.name + "_link"
            store_data(newvarname, data={'x': xnew, 'y': f(xnew)})
        elif method == 'cubic':
            f2 = interp1d(x, y, kind='cubic')
            newvarname = link + "_" + self.name + "_link"
            store_data(newvarname, data={'x': xnew, 'y': f2(xnew)})
        elif method == 'quad_spline':
            tck = interpolate.splrep(x, y, s=0, k=2)
            ynew = interpolate.splev(xnew, tck, der=0)
            newvarname = link + "_" + self.name + "_link"
            store_data(newvarname, data={'x': xnew, 'y': ynew})

        else:
            print('Error: choose interpolation method.')
            print('linear, cubic, quad_spline')
            return

        self.links[name] = newvarname


# Global Variables
hover_time = HoverTime()
data_quants = OrderedDict()
interactive_window = None  # 2D interactive window that appears whenever plotting spectrograms w/ tplot.
# If option 't_average' is set by user, then x and y values on this plot are the average of the user-specified
# number of seconds for which the cursor location should be averaged.
static_window = None  # 2D window showing data at certain point in time from a spectrogram plot.
static_tavg_window = None  # 2D window showing averaged y and z data for a specified time range from a spectrogram plot.
tplot_opt_glob = dict(tools="xpan,crosshair,reset",
                      min_border_top=15, min_border_bottom=0,
                      title_align='center', window_size=[800, 800],
                      title_size='12pt', title_text='', crosshair=True)
lim_info = {}
extra_layouts = {}

if using_graphics:
    pytplotWindows = []  # This is a list that will hold future qt windows
    from . import QtPlotter

    qt_plotters = {'qtTVarFigure1D': QtPlotter.TVarFigure1D,
                   'qtTVarFigureSpec': QtPlotter.TVarFigureSpec,
                   'qtTVarFigureAlt': QtPlotter.TVarFigureAlt,
                   'qtTVarFigureMap': QtPlotter.TVarFigureMap}

bokeh_plotters = {'bkTVarFigure1D': HTMLPlotter.TVarFigure1D,
                  'bkTVarFigureMap': HTMLPlotter.TVarFigureMap,
                  'bkTVarFigureAlt': HTMLPlotter.TVarFigureAlt,
                  'bkTVarFigureSpec': HTMLPlotter.TVarFigureSpec}

from .store_data import store_data
from .tplot import tplot
from .get_data import get_data
from .xlim import xlim
from .ylim import ylim
from .zlim import zlim
from .tlimit import tlimit
from .tplot_save import tplot_save
from .tplot_names import tplot_names
from .tplot_restore import tplot_restore
from .get_timespan import get_timespan
from .tplot_options import tplot_options
from .tplot_rename import tplot_rename
from .get_ylimits import get_ylimits
from .timebar import timebar
from .del_data import del_data
from .timespan import timespan
from .options import options
from .timestamp import timestamp
from .cdf_to_tplot import cdf_to_tplot
from .netcdf_to_tplot import netcdf_to_tplot
from .tplot_utilities import compare_versions
from .link import link
from pytplot.tplot_math import *

# If we are in an ipython environment, set the gui to be qt5
# This allows the user to interact with the window in real time
try:
    magic = get_ipython().magic
    magic(u'%gui qt5')
except:
    pass

# from module import all files
if using_graphics:
    pg.mkQApp()

compare_versions()



import pytplot
import numpy as np


#AVERAGE AT RESOLUTION
#take average of column over discrete periods of time
def avg_res_data(tvar1,res,new_tvar='tvar_avg_res'):
    time_varying = pytplot.data_quants[tvar1].spec_bins_time_varying
    #grab info from tvar
    df = pytplot.data_quants[tvar1].data.copy()
    if (pytplot.data_quants[tvar1].spec_bins is not None) and (time_varying == True):
        dfspec = pytplot.data_quants[tvar1].spec_bins.copy()
    time = df.index
    start_t = df.index[0]
    end_t = df.index[-1]
    df_index = list(df.columns)
    #create list of times spanning tvar range @ specified res
    res_time = np.arange(start_t,end_t+1,res)
    new_res_time = np.array([])
    #find closest time to new resolution times
    for t in res_time:
        if t not in time:
            tdiff = abs(time-t)
            new_res_time = np.append(new_res_time,time[tdiff.argmin()])
        else:
            new_res_time = np.append(new_res_time,t)
    #make sure no duplicate times from resolution time rounding
    new_res_time = np.unique(new_res_time)
    #shift start time array
    start_t = np.roll(new_res_time,1)
    end_t = new_res_time
    start_t = np.delete(start_t,0)
    end_t = np.delete(end_t,0)
    #initialize arrays
    avg_bin_data = []
    avg_bin_spec = []
    avg_bin_time = np.array([])
    #for each time bin
    for it,t in enumerate(start_t):
        #for each data column
        data_avg_bin = np.array([])
        spec_avg_bin = np.array([])
        for i in df_index:
            #append localized bin average to data_avg_bin
            data_avg_bin = np.append(data_avg_bin,[(df.loc[start_t[it]:end_t[it]])[i].mean()])
            if (pytplot.data_quants[tvar1].spec_bins is not None) and (time_varying == True):
                spec_avg_bin = np.append(spec_avg_bin,[(dfspec.loc[start_t[it]:end_t[it]])[i].mean()])
        #append whole array of bin averages (over n columns) to avg_bin_data
        avg_bin_data = avg_bin_data + [data_avg_bin.tolist()]
        if (pytplot.data_quants[tvar1].spec_bins is not None) and (time_varying == True):
            avg_bin_spec = avg_bin_spec + [spec_avg_bin.tolist()]
        avg_bin_time = np.append(avg_bin_time,t)
    #store data in new_tvar
    if (pytplot.data_quants[tvar1].spec_bins is not None) and (time_varying == True):
        pytplot.store_data(new_tvar, data={'x':avg_bin_time,'y':avg_bin_data,'v':avg_bin_spec})
    elif pytplot.data_quants[tvar1].spec_bins is not None:
        pytplot.store_data(new_tvar, data={'x':avg_bin_time,'y':avg_bin_data,'v':pytplot.data_quants[tvar1].spec_bins})
    else:
        pytplot.store_data(new_tvar, data={'x':avg_bin_time,'y':avg_bin_data})
    return new_tvar    


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np

#DERIVE
#take derivative w.r.t. time, store in new_tvar
def derive(tvar1,new_tvar='tvar_derive'):
    #separate and derive data
    time = pytplot.data_quants[tvar1].data.index
    data1 = pytplot.data_quants[tvar1].data
    df_index = pytplot.data_quants[tvar1].data.columns
    new_df = []
    for i in df_index:
        tv1_col = data1[i]
        data = np.diff(tv1_col)/np.diff(time)
        new_df = new_df + [data]
    new_df = np.transpose((list(new_df)))
    time = np.delete(time,0)
    #store differentiated data
    if (pytplot.data_quants[tvar1].spec_bins is not None) and (pytplot.data_quants[tvar1].spec_bins_time_varying == True):
        pytplot.store_data(new_tvar,data={'x':time, 'y':new_df, 'v':pytplot.data_quants[tvar1].spec_bins})
    else:
        pytplot.store_data(new_tvar,data={'x':time, 'y':new_df})
    return new_tvar


#!/usr/bin/python
# -*- coding: utf-8 -*-

import unittest

from ...mms import mms_load_fgm, mms_load_scm, mms_load_fpi, mms_load_hpca, mms_load_eis, mms_load_feeps, mms_load_edp, mms_load_edi, mms_load_aspoc
from ...utilities.data_exists import data_exists

class FEEPSLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_feeps(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_epd_feeps_srvy_l2_electron_bottom_intensity_sensorid_12'))

class EISLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_eis(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_epd_eis_phxtof_proton_flux_omni'))

class FPILoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_fpi(trange=['2015-10-16/14:00', '2015-10-16/15:00'])
        self.assertTrue(data_exists('mms1_dis_energyspectr_omni_fast'))

class HPCALoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_hpca(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_hpca_hplus_number_density'))

class EDILoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_edi(trange=['2016-10-17/13:00', '2016-10-17/14:00'])
        self.assertTrue(data_exists('mms1_edi_e_gse_srvy_l2'))

class ASPOCLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_aspoc(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_aspoc_ionc'))

class EDPLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_edp(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))

    def test_load_multiple_sc(self):
        data = mms_load_edp(probe=['1', '2', '3', '4'], trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_edp_dce_gse_fast_l2'))
        self.assertTrue(data_exists('mms2_edp_dce_gse_fast_l2'))
        self.assertTrue(data_exists('mms3_edp_dce_gse_fast_l2'))
        self.assertTrue(data_exists('mms4_edp_dce_gse_fast_l2'))

    def test_load_brst_data(self):
        data = mms_load_edp(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_edp_dce_gse_brst_l2'))

class FGMLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_fgm(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))

    def test_load_multiple_sc(self):
        data = mms_load_fgm(probe=['1', '2', '3', '4'], trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_srvy_l2'))
        self.assertTrue(data_exists('mms2_fgm_b_gse_srvy_l2'))
        self.assertTrue(data_exists('mms3_fgm_b_gse_srvy_l2'))
        self.assertTrue(data_exists('mms4_fgm_b_gse_srvy_l2'))

    def test_load_brst_data(self):
        data = mms_load_fgm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_fgm_b_gse_brst_l2'))

class SCMLoadTestCases(unittest.TestCase):
    def test_load_default_data(self):
        data = mms_load_scm(trange=['2015-10-16', '2015-10-16/01:00'])
        self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2'))

    # def test_load_multiple_sc(self):
    #     data = mms_load_scm(probe=['1', '2', '3', '4'], trange=['2017-12-15', '2017-12-16'])
    #     self.assertTrue(data_exists('mms1_scm_acb_gse_scsrvy_srvy_l2'))
    #     self.assertTrue(data_exists('mms2_scm_acb_gse_scsrvy_srvy_l2'))
    #     self.assertTrue(data_exists('mms3_scm_acb_gse_scsrvy_srvy_l2'))
    #     self.assertTrue(data_exists('mms4_scm_acb_gse_scsrvy_srvy_l2'))

    def test_load_brst_data(self):
        data = mms_load_scm(data_rate='brst', trange=['2015-10-16/13:06', '2015-10-16/13:10'])
        self.assertTrue(data_exists('mms1_scm_acb_gse_scb_brst_l2'))

if __name__ == '__main__':
    unittest.main()


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

from bokeh.core.properties import String
from bokeh.models.annotations import ColorBar
#from bokeh.core.properties import String, Instance
#from bokeh.models import LayoutDOM, Slider

#This is the start of a future implementation of JS_CODE
#when bokeh moves to typescript.  Please ignore for now.  
JS_CODE2 = '''
import {min, max, range} from "core/util/array"
 
import {ColorBar, ColorBarView} from "models/annotations/color_bar"
import {isEmpty} from "core/util/object"

SHORT_DIM = 25
LONG_DIM_MIN_SCALAR = 0.3
LONG_DIM_MAX_SCALAR = 0.8
export class ColorBarSideTitleView extends ColorBarView
 
  compute_legend_dimensions () {
    let legend_height, legend_width;
    const image_dimensions = this.model._computed_image_dimensions();
    const [image_height, image_width] = [image_dimensions.height, image_dimensions.width];
 
    const label_extent = this._get_label_extent();
    const title_extent = this.model._title_extent();
    const tick_extent = this.model._tick_extent();
    const { padding } = this.model.padding;
 
    switch (this.model.orientation) {
      case "vertical":
        legend_height = image_height;
        legend_width = image_width + tick_extent + title_extent + label_extent + (padding * 2);
        break;
      case "horizontal":
        legend_height = image_height + tick_extent + label_extent + (padding * 2);
        legend_width = image_width + title_extent + (padding * 2);
        break;
    }
    
    return {height: legend_height, width: legend_width}
   
  }
   
  render() {
    if (!this.model.visible || (this.model.color_mapper == null)) {
      return
    }
    const { ctx } = this.plot_view.canvas_view;
    ctx.save()
 
    const {sx, sy} = this.compute_legend_location();
    ctx.translate(location.sx, location.sy);
    this._draw_bbox(ctx);
 
    const image_offset = this._get_image_offset();
    ctx.translate(image_offset.x, image_offset.y);
 
    this._draw_image(ctx);
 
    if ((this.model.color_mapper.low != null) && (this.model.color_mapper.high != null)) {
      const tick_info = this.model.tick_info();
      this._draw_major_ticks(ctx, tick_info);
      this._draw_minor_ticks(ctx, tick_info);
      this._draw_major_labels(ctx, tick_info);
 
    if (this.model.title) {
      this._draw_title(ctx);
    }
    ctx.restore()
  }
 
  _draw_title: (ctx) ->
    if not @visuals.title_text.doit
      return
    image = @compute_legend_dimensions()
    ctx.save()
    @visuals.title_text.set_value(ctx)
    ctx.rotate(-Math.PI/2)
    ctx.textAlign="center"
    ctx.fillText(@model.title, -image.height * 0.5, image.width-@model._title_extent())
    ctx.restore()
 
  _get_image_offset: () ->
    # Returns image offset relative to legend bounding box
    x = @model.padding
    y = 0
    return {x: x, y: y}
 
  _get_label_extent: () ->
    major_labels = @model.tick_info().labels.major
    if @model.color_mapper.low? and @model.color_mapper.high? and not isEmpty(major_labels)
      ctx = @plot_view.canvas_view.ctx
      ctx.save()
      @visuals.major_label_text.set_value(ctx)
 
      switch @model.orientation
        when "vertical"
          label_extent = max((ctx.measureText(label.toString()).width for label in major_labels))
        when "horizontal"
          label_extent = text_util.get_text_height(@visuals.major_label_text.font_value()).height
 
      label_extent += @model.label_standoff
      ctx.restore()
    else
      label_extent = 0
    return label_extent
 
export class ColorBarSideTitle extends ColorBar
  default_view: ColorBarSideTitleView
  type: 'ColorBarSideTitle'
 
  _computed_image_dimensions: () ->
    ###
    Heuristics to determine ColorBar image dimensions if set to "auto"
    Note: Returns the height/width values for the ColorBar's scale image, not
    the dimensions of the entire ColorBar.
    If the short dimension (the width of a vertical bar or height of a
    horizontal bar) is set to "auto", the resulting dimension will be set to
    25 px.
    For a ColorBar in a side panel with the long dimension (the height of a
    vertical bar or width of a horizontal bar) set to "auto", the
    resulting dimension will be as long as the adjacent frame edge, so that the
    bar "fits" to the plot.
    For a ColorBar in the plot frame with the long dimension set to "auto", the
    resulting dimension will be the greater of:
      * The length of the color palette * 25px
      * The parallel frame dimension * 0.30
        (i.e the frame height for a vertical ColorBar)
    But not greater than:
      * The parallel frame dimension * 0.80
    ###

    frame_height = @plot.plot_canvas.frame._height.value
    frame_width = @plot.plot_canvas.frame._width.value
    title_extent = 0

    switch @orientation
      when "vertical"
        if @height == 'auto'
          if @panel?
            height = frame_height
          else
            height = max([@color_mapper.palette.length * SHORT_DIM,
                            frame_height * LONG_DIM_MIN_SCALAR])
            height = min([height,
                            frame_height * LONG_DIM_MAX_SCALAR - 2 * @padding - title_extent])
        else
          height = @height
 
        width = if @width == 'auto' then SHORT_DIM else @width

      when "horizontal"
        height = if @height == 'auto' then SHORT_DIM else @height
 
        if @width == 'auto'
          if @panel?
            width = frame_width
          else
            width = max([@color_mapper.palette.length * SHORT_DIM,
                           frame_width * LONG_DIM_MIN_SCALAR])
            width = min([width,
                           frame_width * LONG_DIM_MAX_SCALAR - 2 * @padding])
        else
          width = @width

    return {"height": height, "width": width}
     
 
 
'''

JS_CODE = '''
import {min, max} from "core/util/array"
 
import {ColorBar, ColorBarView} from "models/annotations/color_bar"
import {isEmpty} from "core/util/object"

SHORT_DIM = 25
LONG_DIM_MIN_SCALAR = 0.3
LONG_DIM_MAX_SCALAR = 0.8
export class ColorBarSideTitleView extends ColorBarView
 
  compute_legend_dimensions: () ->
    image_dimensions = @model._computed_image_dimensions()
    [image_height, image_width] = [image_dimensions.height, image_dimensions.width]
 
    label_extent = @_get_label_extent()
    title_extent = @model._title_extent()
    tick_extent = @model._tick_extent()
    padding = @model.padding
 
    switch @model.orientation 
      when "vertical"
        legend_height = image_height
        legend_width = image_width + tick_extent + title_extent + label_extent + (padding * 2)
        break
      when "horizontal"
        legend_height = image_height + tick_extent + label_extent + (padding * 2)
        legend_width = image_width + title_extent + (padding * 2)
        break
    
    
    return {height: legend_height, width: legend_width}
   
   
  render: () -> 
    if not @model.visible or not @model.color_mapper?
      return

    ctx = @plot_view.canvas_view.ctx
    ctx.save()
 
    {sx, sy} = @compute_legend_location()
    ctx.translate(sx, sy)
    @_draw_bbox(ctx)
 
    image_offset = @_get_image_offset()
    ctx.translate(image_offset.x, image_offset.y)
 
    @_draw_image(ctx)
 
    if @model.color_mapper.low? and @model.color_mapper.high?
      tick_info = @model.tick_info()
      @_draw_major_ticks(ctx, tick_info)
      @_draw_minor_ticks(ctx, tick_info)
      @_draw_major_labels(ctx, tick_info)

    if @model.title
      @_draw_title(ctx)
    ctx.restore()
 
  _draw_title: (ctx) ->
    if not @visuals.title_text.doit
      return
    image = @compute_legend_dimensions()
    ctx.save()
    @visuals.title_text.set_value(ctx)
    ctx.rotate(-Math.PI/2)
    ctx.textAlign="center"
    ctx.fillText(@model.title, -image.height * 0.5, image.width-@model._title_extent())
    ctx.restore()
 
  _get_image_offset: () ->
    # Returns image offset relative to legend bounding box
    x = @model.padding
    y = 0
    return {x: x, y: y}
 
  _get_label_extent: () ->
    major_labels = @model.tick_info().labels.major
    if @model.color_mapper.low? and @model.color_mapper.high? and not isEmpty(major_labels)
      ctx = @plot_view.canvas_view.ctx
      ctx.save()
      @visuals.major_label_text.set_value(ctx)
 
      switch @model.orientation
        when "vertical"
          label_extent = max((ctx.measureText(label.toString()).width for label in major_labels))
        when "horizontal"
          label_extent = text_util.get_text_height(@visuals.major_label_text.font_value()).height
 
      label_extent += @model.label_standoff
      ctx.restore()
    else
      label_extent = 0
    return label_extent
 
export class ColorBarSideTitle extends ColorBar
  default_view: ColorBarSideTitleView
  type: 'ColorBarSideTitle'
 
  _computed_image_dimensions: () ->
    ###
    Heuristics to determine ColorBar image dimensions if set to "auto"
    Note: Returns the height/width values for the ColorBar's scale image, not
    the dimensions of the entire ColorBar.
    If the short dimension (the width of a vertical bar or height of a
    horizontal bar) is set to "auto", the resulting dimension will be set to
    25 px.
    For a ColorBar in a side panel with the long dimension (the height of a
    vertical bar or width of a horizontal bar) set to "auto", the
    resulting dimension will be as long as the adjacent frame edge, so that the
    bar "fits" to the plot.
    For a ColorBar in the plot frame with the long dimension set to "auto", the
    resulting dimension will be the greater of:
      * The length of the color palette * 25px
      * The parallel frame dimension * 0.30
        (i.e the frame height for a vertical ColorBar)
    But not greater than:
      * The parallel frame dimension * 0.80
    ###

    frame_height = @plot.plot_canvas.frame._height.value
    frame_width = @plot.plot_canvas.frame._width.value
    title_extent = 0

    switch @orientation
      when "vertical"
        if @height == 'auto'
          if @panel?
            height = frame_height
          else
            height = max([@color_mapper.palette.length * SHORT_DIM,
                            frame_height * LONG_DIM_MIN_SCALAR])
            height = min([height,
                            frame_height * LONG_DIM_MAX_SCALAR - 2 * @padding - title_extent])
        else
          height = @height
 
        width = if @width == 'auto' then SHORT_DIM else @width

      when "horizontal"
        height = if @height == 'auto' then SHORT_DIM else @height
 
        if @width == 'auto'
          if @panel?
            width = frame_width
          else
            width = max([@color_mapper.palette.length * SHORT_DIM,
                           frame_width * LONG_DIM_MIN_SCALAR])
            width = min([width,
                           frame_width * LONG_DIM_MAX_SCALAR - 2 * @padding])
        else
          width = @width

    return {"height": height, "width": width}
     
 
 
'''
class ColorBarSideTitle(ColorBar):
    #__javascript__ = ["https://cdnjs.cloudflare.com/ajax/libs/underscore.js/1.8.3/underscore-min.js"]
    __implementation__ = JS_CODE 


import pytplot
import pandas as pd

#SPEC BIN MULTIPLICATION
#multiply spec_bin values by tvar data, store in new_tvar
def spec_mult(tvar1,new_tvar='tvar_specmult'):
    dataframe = pytplot.data_quants[tvar1].data
    specframe = pytplot.data_quants[tvar1].spec_bins
    new_df = pd.DataFrame(dataframe.values*specframe.values, columns=dataframe.columns, index=dataframe.index)    
    pytplot.store_data(new_tvar,data={'x':new_df.index,'y':new_df.values})
    return new_tvar


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pytplot
import pydivide
import scipy
import numpy as np
import scipy.fftpack
import matplotlib.pyplot as plt
from astropy.convolution import convolve,Box1DKernel

insitu = pydivide.read('2017-06-19')
t = insitu['Time']
data = insitu['SPACECRAFT']['ALTITUDE']
pytplot.store_data('sgx',data = {'x':t, 'y':data})
 
N= len(pytplot.data_quants['sgx'].data.index)
x = pytplot.data_quants['sgx'].data.index
y = pytplot.data_quants['sgx'].data
print(N)
print(x)
print(y)
def fourier_smooth():
    N= len(pytplot.data_quants['sgx'].data.index)
    x = pytplot.data_quants['sgx'].data.index
    y = pytplot.data_quants['sgx'].data
    w = scipy.fftpack.rfft(y)
    f = scipy.fftpack.rfftfreq(N, x[1]-x[0])
    w2 = w.copy()
    y2 = scipy.fftpack.irfft(w2)
    t = plt.plot(x,y,'r')
    s = plt.plot(x,y2)
    plt.show([s,t])

def boxcar_smooth():
    N= len(pytplot.data_quants['sgx'].data.index)
    x = pytplot.data_quants['sgx'].data.index
    y = pytplot.data_quants['sgx'].data
    print((y.values))
    smoothed = convolve(y.values,Box1DKernel(4))
    s = plt.plot(x,smoothed)
    t = plt.plot(x,y,'r')
    plt.show([s,t])

boxcar_smooth()





# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np
from scipy import interpolate
from scipy.interpolate import interp1d

#TVAR INTERPOLATION
#interpolate tvar2 to tvar1 cadence
def interpolate(tvar1,tvar2,interp='linear'):
    #crop data
    tv1_t,tv1_d,tv2_t,tv2_d = pytplot.crop(tvar1,tvar2)
    df_index = pytplot.data_quants[tvar1].data.columns
    #interpolate to tvar1 cadence
    if interp == 'linear':
        name1 = tvar1 + "_interp"
        name2 = tvar2 + "_interp"
        new_df = []
        for i in df_index:
            tv2_col = [item[i] for item in tv2_d]
            f = interp1d(tv2_t,tv2_col,fill_value="extrapolate")
            new_df = new_df + [f(tv1_t)]
        new_df = np.transpose((list(new_df)))
        #store interpolated tvars as 'X_interp'
        pytplot.store_data(name1, data={'x':tv1_t,'y':tv1_d})
        pytplot.store_data(name2, data={'x':tv1_t,'y':new_df})
    elif interp == 'cubic':
        new_df = []
        for i in df_index:
            tv2_col = [item[i] for item in tv2_d]
            f = interp1d(tv2_t,tv2_col,kind='cubic',fill_value='extrapolate')
            new_df = new_df + [f(tv1_t)]
        new_df = np.transpose((list(new_df)))
        name1 = tvar1 + "_interp"
        name2 = tvar2 + "_interp"
        #store interpolated tvars as 'X_interp'
        pytplot.store_data(name1, data={'x':tv1_t,'y':tv1_d})
        pytplot.store_data(name2, data={'x':tv1_t,'y':new_df})
    elif interp == 'quad_spline':
        new_df = []
        for i in df_index:
            tv2_col = [item[i] for item in tv2_d]
            tck = interpolate.splrep(tv2_t, tv2_col, s=0,k=2)
            ynew = interpolate.splev(tv1_t, tck, der=0)
            new_df = new_df + [ynew]
        new_df = np.transpose((list(new_df)))
        name1 = tvar1 + "_interp"
        name2 = tvar2 + "_interp"
        #store interpolated tvars as 'X_interp'
        pytplot.store_data(name1, data={'x':tv1_t,'y':tv1_d})
        pytplot.store_data(name2, data={'x':tv1_t,'y':new_df})
    return name1,name2


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

from bokeh.core.properties import String
from bokeh.models import LayoutDOM
import os 
import datetime

import pytplot

JS_CODE = '''
import * as p from "core/properties"
import {div, empty} from "core/dom"
import {LayoutDOM, LayoutDOMView} from "models/layouts/layout_dom"

export class TimeStampView extends LayoutDOMView

    initialize: (options) ->
        super(options)
        
        @render()
        
    render: () ->
        empty(@el)
        @el.appendChild(div({
          style: {
            'width': '800px'
            'word-spacing': '10px'
            'font-size': '11px'
            'color': '#000000'
            'background-color': '#ffffff'
            }
        }, "#{ @model.text }"))

export class TimeStamp extends LayoutDOM

    default_view: TimeStampView
    
    type: "TimeStamp"
    
    @define {
        text: [ p.String ]
    }
 


'''


class TimeStamp(LayoutDOM):
    __implementation__ = JS_CODE
    text = String(default = "Testing")




import numpy as np
import pytplot
import pydivide
from scipy import signal

insitu,iuvs = pydivide.read('2017-06-19','2017-06-20')
t = insitu['Time']
data = insitu['SPACECRAFT']['ALTITUDE']
pytplot.store_data('sgx',data = {'x':t, 'y':data})

def pwr_spec(tvar,nbp=256,nsp=128):
    x = pytplot.data_quants[tvar].data.index.values
    y = pytplot.data_quants[tvar].data.values
    
    l = len(x)
    shift_lsp = np.arange(0,l-1,nsp)
    #print(shift_lsp)
    #print(x[0:255])
    for i in shift_lsp-nsp:
        x_n = x[i:i+nsp-1]
        y_n = y[i:i+nsp-1]
    
        x_n = x_n - x_n[0]
        c = np.polyfit(x_n,y_n,1)
        #print(c[0]*x_n[0:nbp-1])
        y_trend = c[0]*x_n + c[1]
        y_n = y_n - y_trend
        print(y_n)
    
    w = signal.get_window("hanning",nbp)
    f,pxx = signal.periodogram(y[0:nbp], window = w, detrend = lambda x: x)
    
pwr_spec('sgx')
    


'''

 Purpose:
    this function returns the FEEPS active eyes,
    based on date/probe/species/rate

 Output:
    Returns a hash table containing 2 hash tables:
       output['top'] -> maps to the active top eyes
       output['bottom'] -> maps to the active bottom eyes
 
 Notes:
 1) Burst mode should include all sensors (TOP and BOTTOM):
     electrons: [1, 2, 3, 4, 5, 9, 10, 11, 12]
     ions: [6, 7, 8]
 
 2) SITL should return (TOP only):
     electrons: set_intersection([5, 11, 12], active_eyes)
     ions: None
     
 3) From Drew Turner, 9/7/2017, srvy mode:
 
   - before 16 August 2017:
      electrons: [3, 4, 5, 11, 12]
      iond: [6, 7, 8]
 
   - after 16 August 2017:
       MMS1
         Top Eyes: 3, 5, 6, 7, 8, 9, 10, 12
         Bot Eyes: 2, 4, 5, 6, 7, 8, 9, 10

       MMS2
         Top Eyes: 1, 2, 3, 5, 6, 8, 10, 11
         Bot Eyes: 1, 4, 5, 6, 7, 8, 9, 11

       MMS3
         Top Eyes: 3, 5, 6, 7, 8, 9, 10, 12
         Bot Eyes: 1, 2, 3, 6, 7, 8, 9, 10

       MMS4
         Top Eyes: 3, 4, 5, 6, 8, 9, 10, 11
         Bot Eyes: 3, 5, 6, 7, 8, 9, 10, 12

'''

def mms_feeps_active_eyes(trange, probe, data_rate, species, level):
    


# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pytplot
import numpy as np

def tplot_ascii(tvar,filename='tvar_data'):
    #grab data, prepend index column
    df = pytplot.data_quants[tvar].data.reset_index()
    sbf = pytplot.data_quants[tvar].spec_bins.reset_index()
    #save data and spec_bins
    sb_name = filename + '_v'
    np.savetxt(filename, df, fmt="%-14s", delimiter = '   ')
    np.savetxt(sb_name, sbf, fmt="%-14s", delimiter = '   ')
    
